import pandas as pd
import numpy as np
import os
import ncls
import re
import gzip
import logging
from time import time
from collections import defaultdict
from dataclasses import dataclass
from ctypes import c_int
import typing
import scbamtools.util as util


## GTF I/O
def attr_to_dict(attr_str: str) -> dict:
    """
    Helper for reading the GTF. Parsers the attr_str found in the GTF
    attribute column (col 9). Returns a dictionary from the key, value
    pairs.
    """
    d = {}
    for match in re.finditer(r"(\w+) \"(\S+)\";", attr_str):
        key, value = match.groups()
        d[key] = value

    if not "gene_name" in d:
        d["gene_name"] = d.get("gene_id", d.get("transcript_id", "na"))

    return d


def load_GTF(
    src: typing.Union[str, typing.TextIO],
    attributes: list = ["transcript_id", "transcript_type", "gene_name"],
    features: list = ["exon", "CDS", "UTR", "transcript"],
) -> pd.DataFrame:
    """
    Loads GTF-formatted annotation data (such as GENCODE) and converts into a tabular DataFrame. Coordinates will be converted
    to zero-based, half-open interval (C, python, BED,...).

    Args:
        src (filename or file-like object): Source of GTF annotation records. Can be gzip compressed (filename ending in .gz)
        attributes (list, optional): Which attributes to extract from the attribute column. Defaults to ["transcript_id", "transcript_type", "gene_name"].
        features (list, optional): which GTF features to care about. Defaults to ["exon", "CDS", "UTR", "transcript"].

    Returns:
        pandas.DataFrame: all GTF annotation data converted to a DataFrame with columns: chrom, feature, start (0-based), end, strand, [attributes]
    """

    ## Efficient representation of GTF features as bit-mask
    feat2bit = {
        "transcript": 1,
        "exon": 2,
        "UTR": 4,
        "CDS": 8,
    }

    # IDEA: plus (low nibble) and minus strand (high nibble) sense would both fit into one byte!
    # extracting only specific-strand features:
    #
    #   f_plus = mask & 0b1111
    #   f_minus = mask >> 4

    if type(src) is str:
        if src.endswith(".gz"):
            src = gzip.open(src, "rt")
        else:
            src = open(src, "rt")

    fset = set(features)
    data = []
    for line in src:
        if line.startswith("#"):
            continue

        parts = line.split("\t")
        if len(parts) != 9:
            # not a GTF/GFF formatted line
            continue

        chrom, source, feature, start, end, score, strand, frame, attr_str = parts
        if feature not in fset:
            continue

        attrs = attr_to_dict(attr_str)
        row = [chrom, feature, int(start) - 1, int(end), strand] + [
            attrs[a] for a in attributes
        ]
        data.append(row)

    df = pd.DataFrame(
        data, columns=["chrom", "feature", "start", "end", "strand"] + attributes
    ).drop_duplicates()

    def bitflags(columns):
        strand, feature = columns
        flag = feat2bit[feature]
        if strand == "-":
            # features on minus strand use the high-nibble of the bit-flags
            flag = flag << 4

        return flag

    df["feature_flag"] = df[["strand", "feature"]].apply(bitflags, axis=1)
    df["transcript_type"] = df["transcript_type"].apply(
        lambda t: abbreviations.get(t, t)
    )
    return df


def make_lookup(d):
    return np.array([d.get(i, "?") for i in np.arange(256)], dtype=object)


default_lookup = make_lookup(
    {
        # CDS UTR exon transcript-> gm, gf tag values
        0b1011: "C",  # CDS exon
        # This should not arise! Unless, isoforms-are merged
        0b1111: "CU",  # CDS and UTR exon
        0b0111: "U",  # UTR exon
        0b0011: "N",  # exon of a non-coding transcript
        0b1001: "I",  # intron
        0b0101: "I",
        0b0001: "I",
        0b0000: "-",  # intergenic
        0b10110000: "c",  # CDS exon
        0b11110000: "cu",  # CDS and UTR exon
        0b01110000: "u",  # UTR exon
        0b00110000: "n",  # exon of a non-coding transcript
        0b10010000: "i",  # intron
        0b01010000: "i",
        0b00010000: "i",
        0b00000000: "-",  # intergenic
    }
)
default_strand_translation = str.maketrans("CUNIcuni", "cuniCUNI", "")

# short-hands for common transcript types
abbreviations = {
    "protein_coding": "C",
    "processed_transcript": "N",
    "processed_pseudogene": "P",
    "transcribed_unprocessed_pseudogene": "P",
    "transcribed_processed_pseudogene": "P",
    "transcribed_unitary_pseudogene": "P",
    "lncRNA": "L",
    "lincRNA": "I",
    "nonsense_mediated_decay": "D",
    "retained_intron": "R",
    "miRNA": "M",
    "tRNA": "t",
    "snoRNA": "s",
    "rRNA": "r",
}


## Space-efficient container of aggregating overlap information (isoform resolution)
## TODO: find a better, more compact/faster intermediate layer
@dataclass
class IsoformOverlap:
    gene_name: str = "no_name"
    transcript_type: str = "no_type"
    flags: int = 0


def overlaps_to_tags(isoform_overlaps: dict, flags_lookup=default_lookup) -> tuple:
    """_summary_
    Take a dictionary of transcript_id -> IsoformOverlap as input and populate annotation tuples

    Args:
        isoform_overlaps (dict): _description_
        flags_lookup (_type_, optional): _description_. Defaults to default_lookup.

    Returns:

        tuple: (gn, gf, gt)
            a tuple with tag values encoding the overlapping
            GTF features. Each is a list of strings :

                gn: gene names
                gf: function
                gt: transcript type

                gf uses upper case letters for features on the '+' strand and
                lower case letters for features on the '-' strand.

                For a query against the minus strand, upper and lower case
                need to be swapped in order to get features that align in the
                "sense" direction in upper case and "antisense" features in lower-case.

                For a strand-agnostic query, .upper() is called.
                These case-mangling operations are carried out in get_annotation_tags()

    """
    # PROBLEM:
    # order of tag values is given by order of annotation records, not anything
    # reproducible, such as alphabetical order. This is makes testing and counting
    # un-necessarily hard. Let's try to collect by gene and sort gf,gt on gf for same gene

    # by_gn = defaultdict(set)
    # for transcript_id, info in isoform_overlaps.items():
    #     _gf = default_lookup[info.flags]

    #     # print(f">> {transcript_id} => {info.flags:04b} => {_gm} {_gf}")
    #     by_gn[info.gene_name].add((_gf, info.transcript_type))

    # gn = []
    # gf = []
    # gt = []
    # for n, tags in by_gn.items():
    #     for f,t in sorted(tags):
    #         gn.append(n)
    #         gf.append(f)
    #         gt.append(t)

    tags = set()
    for transcript_id, info in isoform_overlaps.items():
        _gf = default_lookup[info.flags]

        # print(f">> {transcript_id} => {info.flags:04b} => {_gm} {_gf}")
        tags.add((info.gene_name, _gf, info.transcript_type))

    gn = []
    gf = []
    gt = []
    for n, f, t in tags:
        gn.append(n)
        gf.append(f)
        gt.append(t)

    # print(f"overlaps_to_tags {gn} {gf} {gt}")
    return (gn, gf, gt)


# First implementation: uses the GTF features directly. Not very optimized. Mainly used to
# build compiled annotation (see below).
# A Classifier instance needs to be passed to GenomeAnnotation as a parameter to __init__ (see below)
class GTFClassifier:
    """
    _summary_
    A simple classifier. Initizialized with a DataFrame of GTF features, the process() function
    will combine GTF features for each transcript to produce condensed annotations for the
    annotation bam tags.
    To that end, the presence/absence of CDS, UTR, exon, transcript features in the set of DataFrame
    indices passed into process() is compiled into a boolean vector. This vector is translated into
    values for gene-model and function values using a feature_map. No hierarchy is enforced at this
    level. All annotations are reported based on the aggregate isoform info retrieved via the feature
    ids.
    """

    def __init__(
        self,
        df,
        hierarchy=["CDS", "UTR", "non-coding", "intron"],
        flags_lookup=default_lookup,
    ):
        self.flags_lookup = flags_lookup
        self.hierarchy = hierarchy
        self.prio = {}
        for i, h in enumerate(hierarchy):
            self.prio[h] = i

        self.df_columns = df.columns
        self.df = df.values
        self.df_core = df[
            ["transcript_id", "transcript_type", "gene_name", "feature_flag"]
        ].values

    def process(self, ids: typing.Iterable[c_int]) -> dict:
        """
        Gather annotation records identified by DataFrame (array) indices
        by isoform: gene_name, gene_type and overlap-flags

        Args:
            ids (Iterable of ints): indices into the array-converted DataFrame of GTF records

        Returns:
            output of overlaps_to_tags
        """
        # print(f">>> process({ids})")
        isoform_overlaps = defaultdict(IsoformOverlap)
        # iterate over the overlapping GTF features group by transcript_id as we go
        i: c_int
        for i in ids:
            (transcript_id, transcript_type, gene_name, flag) = self.df_core[i]
            info = isoform_overlaps[transcript_id]
            info.gene_name = gene_name
            info.transcript_type = transcript_type
            info.flags |= flag
            # print(f"df entry for id={i}: {self.df_core[i]} flags={info.flags:08b}")

        return overlaps_to_tags(isoform_overlaps)  # isoform_overlaps

    # def preprocess(self, ids: typing.Iterable[c_int]) -> tuple:
    #     """
    #     pre-process a list of features into full annotation for either strand,
    #     as well as intermediate data objects that allow correct merging.
    #     These results will be stored when annotation is "compiled" in order to offer
    #     much faster lookups.

    #     Args:
    #         idx (_type_): _description_
    #     """


## Second iteration. Here, we load already pre-compiled, unique combinations of original GTF
# features as they occur in the genome. The beauty is that each of these "compiled" features now
# has zero overlap with other compiled features and each of them already has the gn, gf, gt values
# pre-evaluated ("compiled"). The only issue arises if a read overlaps multiple of these pre-compiled
# "meta"-features. For this we need the joiner() function which aims to combine two sets of gn, gf, gt
# values in a meaningful way. This is slow, but rare enough to leave a substantial speed boost
# on average, when comparing to using raw GTF features directly.
# of note, this still plugs into the same GenomeAnnotation facility (see below). It just changes the
# meaning and number of the underlying features (from individual GTF records to combinatorial overlaps
# these GTF records as they occur in the genome), and it also changes the processing because we no
# longer deal with DataFrame entries but already with gn, gf, gt tuples.
class CompiledClassifier:
    def __init__(self, df, classifications):
        self.cid_table = df["cid"].values
        self.classifications = np.array(classifications, dtype=object)

    @staticmethod
    def get_filenames(path):
        cdf_path = os.path.join(path, "non_overlapping.csv")
        cclass_path = os.path.join(path, "classifications.npy")

        return cdf_path, cclass_path

    @staticmethod
    def files_exist(path):
        cdf_path, cclass_path = CompiledClassifier.get_filenames(path)
        return os.access(cdf_path, os.R_OK) and os.access(cclass_path, os.R_OK)

    # The only piece of work left is merging multiple pre-classified annotations
    # TODO: ensure deterministic ordering of joined annotation tags
    def joiner(self, cidx):
        # we want to report each combination only once
        # Note: This code path is executed rarely

        seen = set()
        gn = []
        gf = []
        gt = []
        for idx in tuple(cidx):
            ann_tags = self.classifications[self.cid_table[idx]]
            for ann in zip(*ann_tags):
                if not ann in seen:
                    seen.add(ann)
                    gn.append(ann[0])
                    gf.append(ann[1])
                    gt.append(ann[2])

        return (gn, gf, gt)

    def process(self, cidx, check_contained=False):
        # fast path
        gene_contains = defaultdict(int)

        if check_contained:
            cidx, start_included, end_included = cidx
            for idx, si, ei in zip(cidx, start_included, end_included):
                c = self.classifications[self.cid_table[idx]]
                for gene in c[0]:
                    gene_contains[gene] |= si
                    gene_contains[gene] |= 2 * ei

        if len(cidx) == 1:
            # fast path
            idx = cidx.pop()
            ann = self.classifications[self.cid_table[idx]]
        else:
            ann = self.joiner(cidx)

        if check_contained:
            for gene, fcontains in gene_contains.items():
                if fcontains < 3:
                    ann[0].append(gene)
                    ann[1].append("-")
                    ann[2].append("-")

        return ann


## Helper functions for working with NCLS
def query_ncls(nc, x0, x1, check_contained=False):
    """
    report only the target_ids from the overlapping
    (start, end, id) tuples returned by NCLS.find_overlap().
    format is frozenset bc that can be hashed and used as a key.
    TODO: directly hook into NCLSIterator functionality to get rid of this overhead
    """

    if check_contained:
        keys = []
        start_contained = []
        end_contained = []
        for start, end, key in nc.find_overlap(x0, x1):
            start_contained.append(x0 >= start)
            end_contained.append(x1 <= end)
            keys.append(key)

        # print(f"query ({x0} - {x1}) strand ={strand} -> {res}")
        return keys, start_contained, end_contained

    else:
        res = [o[2] for o in nc.find_overlap(x0, x1)]
        return res


def decompose(nc):
    """
    Iterate over the unique start or end coordinates of all the intervals
    in a NCLS, essentially decomposing the original NCLS into non-overlapping
    segments of unique target_id combinations.
    This is used to *compile* an NCLS. Compiled, here means that target_ids
    reference not original GTF features, but enumerated, unique combinations of
    the GTF features as they are positioned and potentially overlap in the
    reference sequence.
    The advantage is that each such combination has to be processed into the
    tagging information only once, and that this processing can be done prior
    to any tagging, reducing the complexity to a simple array lookup with O(1).

    :param nc: The NCLS (nested list) with original genomic features from GTF
    :type nc: NCLS64
    :yield: (start, end, frozenset(target_ids) )
    :rtype: None
    """
    logger = logging.getLogger("scbamtools.annotator.compile")
    logger.debug(
        f"compiling strand into non-overlapping and pre-classified annotations"
    )

    starts, ends, ids = np.array(nc.intervals()).T
    logger.debug(f"retrieved {len(starts)} intervals from ncls")

    breakpoints = np.array(sorted(set(starts) | set(ends)))
    logger.debug(f"identified {len(breakpoints)} breakpoints")

    if not len(starts):
        return

    # since we sorted, the first must be a start position
    last_pos = breakpoints[0]
    last_key = frozenset(query_ncls(nc, last_pos, last_pos + 1))
    for bp in breakpoints[1:]:
        # depending on wether this is another start or
        # end position, the transition point can be off by one nt
        # in either direction. Most robust fix was to scan all 3
        # options until the key changes (while ensuring that end > start)
        for x in [-1, 0, +1]:
            if bp + x <= last_pos:
                continue

            key = frozenset(query_ncls(nc, bp + x, bp + x + 1))
            if key != last_key:
                if len(last_key):
                    # ensure we do not yield the empty set
                    yield last_pos, bp + x, last_key
                last_key = key
                last_pos = bp + x
                break


class GenomeAnnotation:
    """
    NCLS-based lookup of genome annotation. This is the top-level interface for the query of genomic intervals.
    GenomeAnnotation is actually a pretty thin wrapper around NCLS for fast 1d-interval queries. The raw
    query results are just sets of numerical ids. Interpretation of these sets is deferred to a "processor", an
    instance of either GTFClassifier or CompiledClassifier. These take on the task of processing the raw ids
    into tuples of gn, gf, gt.
    Because these two layers are tightly integrated, there are convenience classmethods to construct a fully
    functional GenomeAnnotation instance for the most common scenarios (from GTF, from compiled data, ...).
    """

    logger = logging.getLogger("scbamtools.annotator.GenomeAnnotation")

    def __init__(self, df, classifier, is_compiled=False):
        """
        [summary]

        :param df: Annotation data of either raw GTF features or pre-classified combinations
        :type df: DataFrame with at least the following columns: chrom, strand, start, end
        :param processor: evaluates indices into df and returns annotation
            the indices point into the dataframe to all
            overlapping features
        :type processor: function that takes frozenset(indices) as sole argument
        """
        # self.df = df
        self.classifier = classifier

        # find all unique combinations of chrom + strand
        # strands = df["chrom"].drop_duplicates().values
        self.chroms = []
        self.chrom_ncls_map = {}
        self.empty = []
        # df['sign'] = df['strand'].apply(lambda x: -1 if x == '-' else +1)

        t0 = time()
        for chrom in df["chrom"].drop_duplicates().values:
            d = df.query(f"chrom == '{chrom}'")
            nested_list = ncls.NCLS(d["start"], d["end"], d.index)
            self.logger.debug(
                f"constructed nested_list for {chrom} with {len(d)} features"
            )

            self.chrom_ncls_map[chrom] = nested_list
            self.chroms.append(chrom)

        dt = time() - t0
        self.logger.debug(
            f"constructed nested lists of {len(df)} features on {len(self.chroms)} strands in {dt:.3f}s"
        )
        self.is_compiled = is_compiled

    @classmethod
    def from_compiled_index(cls, path):
        cls.logger.info(f"loading compiled index from '{path}'")

        cdf_path, cclass_path = CompiledClassifier.get_filenames(path)
        t0 = time()
        cdf = pd.read_csv(cdf_path, sep="\t")
        classifications = np.load(cclass_path, allow_pickle=True)
        dt = time() - t0
        cls.logger.debug(
            f"loaded compiled annotation index with {len(cdf)} original GTF feature combinations and {len(classifications)} pre-compiled classifications in {dt:.3f} seconds"
        )
        ## Create a classifier which uses the pre-compiled, non-overlapping combinations
        ## and corresponding pre-computed annotation tags
        ga = cls(cdf, CompiledClassifier(cdf, classifications), is_compiled=True)
        return ga

    @classmethod
    def from_GTF(cls, gtf, df_cache=""):
        cls.logger.info(f"loading GTF from '{gtf}'")
        # load GTF the first time. Need to build compiled annotation
        t0 = time()
        df = load_GTF(gtf)
        dt = time() - t0
        cls.logger.debug(f"loaded {len(df)} GTF records in {dt:.3f} seconds")
        if df_cache:
            df.to_csv(df_cache, sep="\t")

        ## Build NCLS with original GTF features
        ga = cls(df, GTFClassifier(df))
        return ga

    @classmethod
    def from_uncompiled_df(cls, path):
        cls.logger.info(f"loading from uncompiled df '{path}'")
        t0 = time()
        df = pd.read_csv(path, sep="\t")
        dt = time() - t0
        cls.logger.debug(f"loaded {len(df)} tabular records in {dt:.3f} seconds")

        ## Build NCLS with original GTF features
        ga = cls(df, GTFClassifier(df))
        return ga

    # def sanity_check(self, df):
    #     for chrom, nc in self.chrom_ncls_map.items():
    #         starts, ends, idx = np.array(nc.intervals()).T
    #         self.logger.debug(
    #             f"checking {chrom} with starts from {starts.min()}-{starts.max()} and idx from {idx.min()}-{idx.max()}"
    #         )

    #         assert idx.max() < len(df)

    def query_idx(self, chrom, start, end, **kw):
        nested_list = self.chrom_ncls_map.get(chrom, [])
        # print(f"nested list for {chrom} -> {nested_list}")
        if nested_list:
            return query_ncls(nested_list, start, end, **kw)
        else:
            return self.empty

    def query_idx_blocks(self, chrom, blocks, **kw):
        idx = []
        for start, end in blocks:
            idx.extend(self.query_idx(chrom, start, end, **kw))

        return idx

    # TODO: do we ever need query_idx_blocks separate from query?
    def query(self, chrom, start, end, **kw):
        idx = self.query_idx(chrom, start, end, **kw)
        return self.classifier.process(idx, **kw)

    def query_blocks(self, chrom, blocks, **kw):
        idx = self.query_idx_blocks(chrom, blocks, **kw)
        return self.classifier.process(idx, **kw)

    def compile(self, path=""):
        if self.is_compiled:
            raise ValueError(
                "attempting to compile an already compiled annotation. Why?? %-/ "
            )

        self.logger.info(f"compiling to '{path}'...")
        if path:
            path = util.ensure_path(path + "/")

        chroms = []
        # strands = []
        cstarts = []
        cends = []
        cids = []
        cidx = []
        cid_lkup = {}

        t0 = time()
        for chrom, nc in self.chrom_ncls_map.items():
            # chrom = strand_key[:-1]
            # strand = strand_key[-1]
            self.logger.debug(f"decomposing {chrom}")
            # decompose will report combinations of real indices into the df
            # for both strands + and - . The task of properly assigning these is
            # shifted to process(), merge() and later queries.
            # In consequence a single compiled annotation element can include features
            # from both strands!
            for start, end, idx in decompose(nc):
                # print(f"start={start} end={end} idx={idx}")
                chroms.append(chrom)
                # strands.append(strand)
                cstarts.append(start)
                cends.append(end)
                if idx not in cid_lkup:
                    cid_lkup[idx] = len(cid_lkup)
                    cidx.append(idx)

                cids.append(cid_lkup[idx])

        self.logger.debug("done")
        cdf = pd.DataFrame(
            dict(chrom=chroms, start=cstarts, end=cends, cid=cids)  # strand=strands,
        )
        dt = time() - t0
        self.logger.debug(
            f"decomposed original GTF annotation in {dt:.3f} seconds. Pre-classifying all disjunct combinations..."
        )

        # pre-classify all unique feature combinations
        classifications = []
        t0 = time()
        for n, idx in enumerate(cidx):
            res = postprocess_tags(
                *self.classifier.process(idx)
            )  # <- ((gn, gf, gt), isoform_overlaps dict)
            # classifications.append(to_tags(res))
            classifications.append(res)

        ## Turn into np.array to make lookup a super-fast array index operation
        classifications = np.array(classifications, dtype=object)
        dt = time() - t0
        self.logger.debug(
            f"pre-classified {n} feature combinations in {dt:.3f} seconds"
        )

        if path:
            cdf_path, cclass_path = CompiledClassifier.get_filenames(path)
            ## Store the compiled DataFrame and pre-classifications
            t0 = time()
            cdf.to_csv(cdf_path, sep="\t", index=False)
            np.save(cclass_path, classifications)
            dt = time() - t0
            self.logger.debug(f"stored compiled annotation index in {dt:.3f} seconds")

        ## Create a secondary Annotator which uses the non-overlapping combinations
        ## and the pre-classified annotations for the actual tagging
        gc = GenomeAnnotation(
            cdf, CompiledClassifier(cdf, classifications), is_compiled=True
        )

        return gc

    def get_annotation_tags(self, chrom: str, strand: str, blocks):
        def collapse_tag_values(values):
            if len(set(values)) == 1:
                return [
                    values[0],
                ]
            else:
                return values

        gn, gf, gt = self.query_blocks(
            chrom, blocks, check_contained=False
        )  # True is experimental and breaks stuff
        if len(gf):
            gn = ",".join(collapse_tag_values(gn))
            gf = ",".join(gf)
            gt = ",".join(collapse_tag_values(gt))
        else:
            gn = "-"
            gf = "-"
            gt = "-"

        if strand == "-":
            gf = gf.translate(default_strand_translation)
        elif strand == "*":
            gf = gf.upper()

        return gn, gf, gt


def postprocess_tags(gn, gf, gt):
    """
    post-process multiple functional annotations for a single gene,
    as can arise if we have multiple isoforms in a region.
    Assuming that this is called during compilation, the query region
    is a minimal combination of uniquely overlapping features.
    Since we postprocess annotation for each gene separately, some
    combinations can then ONLY arise by isoform differences. This is
    not the same scenario as overlapping distinct functional elements of
    all isoforms. We want:
        'C,I' -> "overlaps exonic regions *and* intronic regions in all known isoforms"
        'C|I' -> "overlaps a region that is coding in some and intronic in other isoforms"
        "C,C|I,I" -> "overlaps a region that is coding in all isoforms, a region that is coding in
        some and intronic in others, and a region that is intronic in all isoforms"
    """
    # print("input", gn, gf, gt)
    from collections import defaultdict

    by_gn = defaultdict(set)
    for n, f, t in zip(gn, gf, gt):
        by_gn[n].add((f, t))

    # print(gf_by_gn)
    # print(gt_by_gn)
    gn = sorted(by_gn.keys())
    # print(gn)
    # print("|".join(sorted(gf_by_gn['B'])))
    gf = []
    gt = []
    for n in gn:
        _gf = set()
        _gt = set()
        for f, t in sorted(by_gn[n]):
            _gf.add(f)
            _gt.add(t)

        gf.append("|".join(sorted(_gf)))
        gt.append("|".join(sorted(_gt)))

    # gf = ["|".join(sorted(by_gn[n])) for n in gn]
    # gt = ["|".join(sorted(gt_by_gn[n])) for n in gn]
    return gn, gf, gt



import pytest
import os
import sys

scbamtools_dir = os.path.dirname(__file__) + "/../"


@pytest.fixture(scope="session")
def test_root(tmp_path_factory):
    tmp = tmp_path_factory.mktemp("root")

    return tmp


def run(cmd, *argc, expect_fail=False):
    sys.argv = [
        f"{cmd}.py",
    ] + list(argc)
    from importlib import import_module

    mod = import_module(f"scbamtools.bin.{cmd}")
    res = mod.cmdline()
    print("got result", res)

    if expect_fail:
        assert isinstance(res, Exception) == True
    else:
        assert isinstance(res, Exception) == False


def ubam(*argc, **kwargs):
    run("uBAM", *argc, **kwargs)


def trim(*argc, **kwargs):
    run("trim", *argc, **kwargs)


def ann(*argc, **kwargs):
    run("ann", *argc, **kwargs)



import numpy as np
import logging
import scbamtools.util as util

logger = logging.getLogger("random")


def wrap(seq, n=80):
    buffer = []
    while seq:
        l = min(n, len(seq))
        buffer.append(seq[:l])
        seq = seq[l:]

    return "\n".join(buffer)


class Genome:
    def __init__(self, seqs):
        self.seqs = seqs
        self.chrom_sizes = {}
        for chrom, seq in self.seqs.items():
            self.chrom_sizes[chrom] = len(seq)

    def get_oriented(self, chrom, start, end, strand):
        seq = self.seqs[chrom][start:end]
        if strand == "-":
            seq = util.rev_comp(seq)

        return seq

    def store_fa(self, fname):
        with open(fname, "w") as fa:
            for chrom, seq in sorted(self.seqs.items()):
                wrapped = wrap(seq)
                fa.write(f">{chrom}\n{wrapped}\n\n")


def random_sequences(l=12, n=1, alphabet=list("ACGT")):
    for i in range(n):
        yield "".join(np.random.choice(alphabet, size=l, replace=True))


cell_barcodes = list(random_sequences(l=12, n=5))
UMIs = list(random_sequences(l=8, n=150))


def make_genome(chrom_sizes):
    genome = {}
    for chrom, size in sorted(chrom_sizes.items()):
        seq = next(random_sequences(l=size, n=1))
        genome[chrom] = seq

    return Genome(genome)


logging.basicConfig(level=logging.INFO)
np.random.seed(13770815)

logger.info("generating genome")
chrom_sizes = {}
for line in open("test_data/simple_chrom_sizes.csv"):
    chrom, size = line.rstrip().split("\t")
    chrom_sizes[chrom] = int(size)

genome = make_genome(chrom_sizes)
genome.store_fa("test_data/simple_genome.fa")

logger.info("sampling cDNA reads")


def sample_reads(mrna, l=90, n=200):
    for i in np.random.randint(0, len(mrna) - l + 1, size=n):
        yield mrna[i : i + l]


def sample_barcodes(i):
    return cell_barcodes[i % len(cell_barcodes)], UMIs[i % len(UMIs)]


def sample_quality(l=50, l_good=80, phred_base=33, min_qual=2, max_qual=40):
    x = np.arange(l)
    m = np.where(x < l_good, max_qual - 2, min_qual)
    M = np.where(x < l_good, max_qual, max_qual - 10)
    Q = np.random.randint(m, high=M + 1, size=l)

    return "".join([chr(q + phred_base) for q in Q])


from scbamtools.gene_model import Transcript

import gzip

with open("test_data/simple_mrnas.fa", "w") as mrna:
    with gzip.open("test_data/simple.reads1.fastq.gz", mode="wt") as fq1:
        with gzip.open("test_data/simple.reads2.fastq.gz", mode="wt") as fq2:
            for tx in Transcript.from_GTF(
                "test_data/simple_annotation.gtf", genome=genome
            ):
                seq = tx.spliced_sequence + "A" * 90
                mrna.write(f">{tx.transcript_id}\n{seq}\n")
                for i, read in enumerate(sample_reads(seq)):
                    qual = sample_quality(len(read))
                    fq2.write(f"@read_{tx.transcript_id}_{i}\n{read}\n+\n{qual}\n")

                    bc, umi = sample_barcodes(i)
                    seq1 = bc + umi + "TT"
                    qual = sample_quality(len(seq1), l_good=10)  # "I" * len(seq1)
                    fq1.write(f"@read_{tx.transcript_id}_{i}\n{seq1}\n+\n{qual}\n")

                seq = tx.unspliced_sequence + "A" * 90
                mrna.write(f">{tx.transcript_id}\n{seq}\n")
                for i, read in enumerate(sample_reads(seq)):
                    qual = sample_quality(len(read))
                    fq2.write(f"@read_{tx.transcript_id}_{i}\n{read}\n+\n{qual}\n")

                    bc, umi = sample_barcodes(i)
                    seq1 = bc + umi + "TT"
                    qual = sample_quality(len(seq1), l_good=10)  # "I" * len(seq1)
                    fq1.write(f"@read_{tx.transcript_id}_{i}\n{seq1}\n+\n{qual}\n")



from scbamtools.contrib import __version__, __author__, __license__, __email__
import numpy as np
import sys
from collections import OrderedDict

import datetime
import os
import scbamtools.util as util
import scbamtools.config as config
from scbamtools.contrib import __version__, __author__, __license__
from scbamtools.quant import DGE, default_channels, sparse_summation
import mrfifo as mf


def cmdline():
    parser = util.make_minimal_parser(
        prog="count",
        description="quantify per-cell gene expression from BAM files by counting into a (sparse) DGE matrix",
        main=main,
    )
    parser.add_argument("--config", default="config.yaml", help="path to config-file")

    parser.add_argument(
        "bam_in",
        help="bam input (default=stdin)",
        default=["/dev/stdin"],
        nargs="+",
    )
    parser.add_argument(
        "--skim",
        help="skim through the BAM by investigating only every <skim>-th record (default=1 off)",
        default=1,
        type=int,
        # nargs="+",
    )
    parser.add_argument(
        "--worker-threads",
        help="number of parallel worker processes. (default=4)",
        default=4,
        type=int,
    )
    parser.add_argument(
        "--input-threads",
        help="number of threads to use for BAM decompression. (default=2)",
        default=2,
        type=int,
    )
    parser.add_argument(
        "--pipe-buffer-size",
        help="total pipe buffer size in MB to allocate via mrfifo for interprocess communication (default=16)",
        default=16,
        type=int,
    )
    parser.add_argument(
        "--flavor",
        help="name of the adapter flavor used to retrieve sequences and parameters from the config.yaml. Can also be a mapping choosing specific flavor for each reference (example: 'first@miRNA,default@genome,chrom@rRNA')",
        default="default",
    )
    parser.add_argument(
        "--cell-bc-allow-list",
        help="[OPTIONAL] a text file with cell barcodes. All barcodes not in the list are ignored.",
        default="",
    )
    parser.add_argument(
        "--output",
        help="directory to store the output h5ad and statistics/marginal counts",
        default="dge",
    )
    parser.add_argument(
        "--dge-out",
        help="filename for the output h5ad",
        default="{args.output}/{args.sample}.h5ad",
    )
    parser.add_argument(
        "--summary-out",
        help="filename for the output summary (sum over all vars/genes)",
        default="{args.output}/{args.sample}.summary.tsv",
    )
    parser.add_argument(
        "--bulk-out",
        help="filename for the output summary (sum over all vars/genes)",
        default="{args.output}/{args.sample}.pseudo_bulk.tsv",
    )
    parser.add_argument(
        "--stats-out",
        help="filename for the statistics/counts output",
        default="{args.output}/{args.sample}.stats.tsv",
    )
    args = parser.parse_args()
    return args


def get_counter_class(classpath):

    if "." in classpath:
        mod, cls = classpath.rsplit(".", 1)
        import importlib

        m = importlib.import_module(mod)
    else:
        m = globals().get(classpath)

    return getattr(m, cls)


def get_config_for_refs(args):
    """
    Parse flavor definitions, filling a dictionary with reference name as key and configuration
    as values.
    The values are kwargs dictionaries, including the counter_class argument already replaced
    with the actual class object ready for instantiation.

    The key '*' is used for default settings.

    For convenience, we also collect the names of all channels known by all counter classes, so
    that the DGE object can be created before any data are parsed.
    """
    flavors = args.quant
    # replace str of counter class name/path with actual class object
    for name, config in flavors.items():
        config["name"] = name

    ref_d = {}
    default = "default"
    for f in args.flavor.split(","):
        if "@" in f:
            ref, name = f.split("@")
            ref_d[ref] = flavors[name]
        else:
            default = f

    ref_d["*"] = flavors[default]

    # collect all channel names that can be expected to be generated
    channels = []
    for ref, config in ref_d.items():
        channels.extend(config.get("channels", default_channels))

    return ref_d, sorted(set(channels))


def DGE_counter(Qin, Qerr, abort_flag, stat_list, args={}, **kw):
    with ExceptionLogging(
        "scbamtools.quant.DGE_counter", Qerr=Qerr, exc_flag=abort_flag
    ) as el:
        # load detailed counter configurations for each reference name
        # plus a default setting ('*' key)
        conf_d, channels = get_config_for_refs(args)

        # prepare the sparse-matrix data collection (for multiple channels in parallel)
        dge = DGE(channels=channels, cell_bc_allowlist=args.cell_bc_allowlist)

        # keep track of which reference contained which gene names
        gene_source = {}

        # iterate over chunks of count_data prepared by the worker processes
        for n_chunk, ref, count_data in queue_iter(Qin, abort_flag):
            for gene, cell, channels in count_data:
                dge.add_read(gene=gene, cell=cell, channels=channels)
                # keep track of the mapping reference origin of every gene
                # (in case we combine multiple BAMs)
                if gene.startswith("mm_") and ref != "genome":
                    print(f"gene {gene} is in {ref} ? RLY?")

                gene_source[gene] = ref

            Qin.task_done()

        el.logger.debug("completed counting. Writing output.")

        sparse_d, obs, var = dge.make_sparse_arrays()
        adata = dge.sparse_arrays_to_adata(sparse_d, obs, var)
        adata.var["reference"] = [gene_source[gene] for gene in var]
        adata.uns["sample_name"] = args.sample
        adata.uns["DGE_info"] = (
            f"created with scbamtools.quant version={__version__} on {datetime.datetime.today().isoformat()}"
        )
        adata.uns["DGE_cmdline"] = sys.argv
        adata.write(args.out_dge.format(args=args))

        ## write out marginal counts across both axes:
        #   summing over obs/cells -> pseudo bulk
        #   summing over genes -> UMI/read distribution
        pname = args.out_bulk.format(args=args)

        data = OrderedDict()

        data["sample_name"] = args.sample
        data["reference"] = [gene_source[gene] for gene in var]
        data["gene"] = var
        for channel, M in sparse_d.items():
            data[channel] = sparse_summation(M)

        import pandas as pd

        df = pd.DataFrame(data).sort_values("counts")
        df.to_csv(pname, sep="\t")

        tname = args.out_summary.format(args=args)

        data = OrderedDict()

        # TODO: add back in once we dropped the Rmd QC sheet scripts
        # data["sample_name"] = args.sample
        data["cell_bc"] = obs
        for channel, M in sparse_d.items():
            data[channel] = sparse_summation(M, axis=1)

        df = pd.DataFrame(data)
        df.to_csv(tname, sep="\t")
        el.logger.debug("shutdown")


def simple_check(input, output="/dev/stdout"):
    n = 0
    for line in input:
        n += 1
    return n


def sam_to_DGE(input, channels=[], cell_bc_allow_list="", config={}, **kw):
    import scbamtools.quant
    import logging

    logger = logging.getLogger("scbamtools.count.sam_to_DGE")
    logger.debug(
        f"using config {config['name']} -> counter_class = {config['counter_class']}"
    )
    counter_class = get_counter_class(
        config.get("counter_class", "scbamtools.quant.DefaultCounter")
    )
    counter = counter_class(**config)

    dge = scbamtools.quant.DGE(channels=channels, cell_bc_allow_list=cell_bc_allow_list)
    for bundle in counter.bundles_from_SAM(input):
        cell, gene, channels = counter.process_bundle(bundle)
        if cell != "-":
            dge.add_read(cell, gene, channels)

    channel_d, obs_names, var_names = dge.make_sparse_arrays()
    return {
        "channel_d": channel_d,
        "obs_names": obs_names,
        "var_names": var_names,
        "counter_stats": counter.stats,
    }


def main(args):
    from scbamtools.quant import DGE
    import anndata
    import pandas as pd
    from time import time

    configs, channels = get_config_for_refs(args)

    logger = util.setup_logging(args, f"scbamtools.count.main")
    logger.debug("startup")

    adata_refs = []
    util.ensure_path(args.output + "/")

    t0 = time()
    stats = []
    for input in args.bam_in:
        ref = os.path.basename(input).split(".")[0]
        logger.info(f"processing alignments from {input} -> ref={ref}")
        if not ref in configs:
            config = configs["*"]
            logger.warning(
                f"no config specified for ref='{ref}' ({sorted(configs.keys())}), falling back to default."
            )
        else:
            config = configs[ref]

        w = (
            mf.Workflow("count", total_pipe_buffer_MB=args.pipe_buffer_size)
            .BAM_reader(
                input=input,
                threads=args.input_threads,
                output=mf.FIFO("input_sam", "w"),
            )
            .distribute(
                func=mf.parts.distribute_by_CB,
                input=mf.FIFO("input_sam", "r"),
                outputs=mf.FIFO("dist_{n}", "w", n=args.worker_threads),
                header_detect_func=mf.util.is_header,
                header_fifo="/dev/null",
            )
            .workers(
                func=sam_to_DGE,
                input=mf.FIFO("dist_{n}", "r"),
                channels=channels,
                cell_bc_allow_list=args.cell_bc_allow_list,
                n=args.worker_threads,
                config=config,
            )
            .run()
        )

        from collections import defaultdict

        counter_stats = defaultdict(int)
        counter_stats["ref"] = ref
        counter_stats["config"] = config["name"]

        # collect the counts and make one AnnData object
        adata_shards = []
        na_shards = []
        for name, res in w.result_dict.items():
            if name.startswith("count.worker"):
                adata = DGE.sparse_arrays_to_adata(
                    res["channel_d"], res["obs_names"], res["var_names"]
                )
                # adata.obs["worker"] = name
                if "NA" in adata.obs_names:
                    na_shards.append(adata["NA"])
                    na_mask = np.array([name == "NA" for name in adata.obs_names])
                    # drop the NA gene (for barcode outside allowlist) from
                    # adata and keep it separate
                    adata = adata[~na_mask].copy()

                adata_shards.append(adata)

                for k, n in res["counter_stats"].items():
                    counter_stats[k] += n

        # compile one adata object for all barcode-prefix shards by concatenating on obs (cell barcodes)
        adata = anndata.concat(
            adata_shards, axis=0, join="outer", merge="unique", uns_merge="first"
        )
        adata.obs_names_make_unique()
        # # need to add up everything that is 'NA-x'
        # na_mask = np.array([o.startswith("NA-") for o in adata.obs_names])
        # na = adata[na_mask].X.sum(axis=0)
        # adata = adata[~na_mask].copy()
        # adata["NA"] = na
        adata.var["reference"] = ref
        adata.var["reference"] = pd.Categorical(adata.var["reference"])
        adata.obs[f"n_{ref}_counts"] = sparse_summation(adata.X, axis=1)
        adata_refs.append(adata)
        # adata.write_h5ad(f"{ref}.h5ad")

        stats.append(counter_stats)

    # compile one adata object for all BAM files by concatenating on vars (genes)
    adata = anndata.concat(
        adata_refs, axis=1, join="outer", merge="unique", uns_merge="first"
    )

    # number of molecules counted for this cell
    adata.obs[f"n_counts"] = sparse_summation(adata.X, axis=1)
    # number of genes detected in this cell
    adata.obs["n_genes"] = sparse_summation(adata.X > 0, axis=1)
    for channel in adata.layers.keys():
        adata.obs[f"n_{channel}"] = sparse_summation(adata.layers[channel], axis=1)
        adata.var[f"n_{channel}"] = sparse_summation(adata.layers[channel], axis=0)

    # total counts for a gene across all cells
    adata.var[f"n_counts"] = sparse_summation(adata.X, axis=0)
    # number of cells with at least one count for the gene
    adata.var[f"n_cells"] = sparse_summation(adata.X > 0, axis=0)

    adata.obs.index.name = "cell_bc"
    adata.var.index.name = "gene"
    adata.write_h5ad(args.dge_out)

    # store marginals
    adata.obs.sort_values("n_counts", ascending=False).to_csv(
        args.summary_out, sep="\t"
    )
    adata.var.sort_values("n_counts", ascending=False).to_csv(args.bulk_out, sep="\t")

    # store counting statistics
    df_stats = pd.DataFrame(stats)
    df_stats["sample"] = args.sample
    df_stats = df_stats[sorted(df_stats.columns)].fillna(0)
    # print(df_stats.T)
    df_stats.to_csv(args.stats_out, sep="\t", index=False)

    dt = time() - t0
    n_total = df_stats["SAM_records_total"].sum()
    logger.info(
        f"finished counting {n_total/1E6:.2f}M SAM records in {dt:.1f} seconds ({n_total/dt/1E6:.3f}M/sec)."
    )
    return adata


if __name__ == "__main__":
    args = cmdline()
    args.func(args)



from fixtures import *


def test_help():
    try:
        trim("--help")
    except SystemExit:
        pass


def test_dropseq():
    trim(
        scbamtools_dir + "test_data/adap_test.bam",
        "--bam-out",
        "test_data/adap_test.out.bam",
        "--flavor",
        "dropseq",
    )



from scbamtools.contrib import __version__, __license__, __author__, __email__
import os
import sys
import logging
from collections import defaultdict
import mrfifo as mf
import scbamtools.util as util
import scbamtools.config as config
from scbamtools.annotation import GenomeAnnotation, CompiledClassifier


class AnnotationStats:
    logger = logging.getLogger("scbamtools.annotator.stats")

    def __init__(
        self,
        mapq=defaultdict(int),
        flag=defaultdict(int),
        gn=defaultdict(int),
        gf=defaultdict(int),
        gt=defaultdict(int),
        **kw,
    ):
        from collections import defaultdict

        self.mapq = mapq
        self.flag = flag
        self.gn = gn
        self.gf = gf
        self.gt = gt

        self.last_qname = None

    def count(self, read, gn_val, gf_val, gt_val):
        # count each read only once even if we have multiple alignments or mate pairs
        if read.qname != self.last_qname:
            self.flag[read.flag] += 1
            self.mapq[read.mapping_quality] += 1
            self.gf[gf_val] += 1
            self.gt[gt_val] += 1
            self.gn[gn_val] += 1

        self.last_qname = read.qname

    def as_dict(self):
        return dict(
            mapq=self.mapq,
            flag=self.flag,
            gn=self.gn,
            gf=self.gf,
            gt=self.gt,
        )

    def save_stats(self, fname, fields=["mapq", "flag", "gt", "gf", "gn"]):
        import pandas as pd

        fname = util.ensure_path(fname)

        data = []
        for name in fields:
            d = getattr(self, name)
            for obs, count in sorted(d.items(), key=lambda x: -x[1]):
                data.append((name, obs, count))

        self.logger.debug(f"writing annotation statistics to '{fname}'")
        df = pd.DataFrame(data, columns=["field", "value", "count"])
        df.to_csv(fname, sep="\t", index=None)
        return df


def blocks_from_cigar(cigar, pos=0):
    import re

    start = pos
    for x, op in re.findall(r"(\d+)(\w)", cigar):
        if op == "M":
            pos += int(x)
        elif op == "D":
            pos += int(x)
        elif op == "N":
            yield (start, pos)
            pos += int(x)
            start = pos

    if pos > start:
        yield (start, pos)


def annotate_SAM(input, output, compiled_annotation):
    """_summary_
    Worker sub-process. Annotate SAM lines using compiled GenomeAnnotation
    and place complete gn, gf, gt string values.

    Args:
        compiled_annotation (str): path to directory with compiled annotation information
    """
    logger = logging.getLogger("scbamtools.annotator.annotate_SAM")
    logger.debug(
        f"reading from '{input}' writing to '{output}' compiled_annotation='{compiled_annotation}'"
    )
    ga = GenomeAnnotation.from_compiled_index(compiled_annotation)

    for sam_line in input:
        sam = sam_line.split("\t")
        flags = int(sam[1])
        is_unmapped = flags & 4
        gn = None
        gf = "-"
        gt = None
        if not is_unmapped:
            chrom = sam[2]
            pos = int(sam[3]) - 1  # SAM uses one-based coordinates!!!
            cigar = sam[5]
            strand = "-" if (flags & 16) else "+"
            gn, gf, gt = ga.get_annotation_tags(
                chrom, strand, blocks_from_cigar(cigar, pos=pos)
            )

        tag_str = f"\tgn:Z:{gn}\tgf:Z:{gf}\tgt:Z:{gt}\n"
        output.write(sam_line.replace("\n", tag_str))


def is_header(line):
    return line.startswith("@")


##############################################################
### Here are the main functions as offered via commandline ###
##############################################################


def main(args):
    """_summary_
    Main function of the 'annotate' command. Create the plumbing for parallel worker processes and a single
    collector/writer process.

    Args:
        args (namespace): the command-line arguments reported from the parser
    """
    logger = util.setup_logging(args, name="scbamtools.cutadapt_bam.main")

    w = (
        mf.Workflow("ann")
        .BAM_reader(input=args.bam_in, threads=args.threads_read)
        .distribute(
            input=mf.FIFO("input_sam", "rt"),
            outputs=mf.FIFO("dist{n}", "wt", n=args.threads_work),
            chunk_size=1,
            header_detect_func=is_header,
            header_broadcast=False,
            header_fifo=mf.FIFO("orig_header", "wt"),
        )
        .workers(
            input=mf.FIFO("dist{n}", "rt"),
            output=mf.FIFO("out{n}", "wt"),
            func=annotate_SAM,
            compiled_annotation=args.compiled,
            n=args.threads_work,
        )
        .add_job(
            func=util.update_header,
            input=mf.FIFO("orig_header", "rt"),
            output=mf.FIFO("new_header", "wt"),
            progname="ann.py",
        )
        .collect(
            inputs=mf.FIFO("out{n}", "rt", n=args.threads_work),
            header_fifo=mf.FIFO("new_header", "rt"),
            output=mf.FIFO("out_sam", "wt"),
            log_rate_every_n=1000000,
            log_rate_template="annotated {M_out:.1f} M SAM records ({mps:.3f} M/s, overall {MPS:.3f} M/s)",
            chunk_size=1,
        )
        .funnel(
            input=mf.FIFO("out_sam", "rt"),
            output=args.bam_out,
            _manage_fifos=False,
            func=mf.parts.bam_writer,
            threads=args.threads_write,
            fmt=f"Sh{args.bam_out_mode}",
        )
        .run()
    )


def build_compiled_annotation(args):
    logger = logging.getLogger("scbamtools.annotator.build_compiled_annotation")
    if CompiledClassifier.files_exist(args.compiled):
        logger.warning(
            "already found a compiled annotation. use --force-overwrite to overwrite"
        )
        # ga = GenomeAnnotation.from_compiled_index(args.compiled)
        if not args.force_overwrite:
            return

    if args.tabular and os.access(args.tabular, os.R_OK):
        ga = GenomeAnnotation.from_uncompiled_df(args.tabular)
    else:
        ga = GenomeAnnotation.from_GTF(args.gtf, df_cache=args.tabular)

    ga = ga.compile(args.compiled)


def query_regions(args):
    logger = logging.getLogger("scbamtools.annotator.query_regions")
    if args.compiled:
        ga = GenomeAnnotation.from_compiled_index(args.compiled)
    else:
        ga = GenomeAnnotation.from_GTF(args.gtf)

    for region in args.region:
        logger.debug(f"querying region '{region}'")
        chrom, coords, strand = region.split(":")
        start, end = coords.split("-")

        gn, gf, gt = ga.get_annotation_tags(
            chrom,
            strand,
            [
                (int(start), int(end)),
            ],
        )
        print(f"gn={gn}\tgf={gf}\tgt={gt}")


def query_gff(args):
    logger = logging.getLogger("scbamtools.annotator.query_gff")
    if args.compiled:
        ga = GenomeAnnotation.from_compiled_index(args.compiled)
    else:
        ga = GenomeAnnotation.from_GTF(args.gtf)

    for line in open(args.gff):
        if line.startswith("#"):
            continue

        parts = line.split("\t")
        if len(parts) != 9:
            # not a GTF/GFF formatted line
            continue

        chrom, source, feature, start, end, score, strand, frame, attr_str = parts

        start = int(start) - 1
        end = int(end)

        logger.debug(f"querying region '{chrom}:{start}-{end}:{strand}'")

        gn, gf, gt = ga.get_annotation_tags(
            chrom,
            strand,
            [
                (start, end),
            ],
        )
        print(f'{line.rstrip()} gn "{gn}"; gf "{gf}"; gt "{gt}";')


def parse_args():
    parser = util.make_minimal_parser("ann.py")  # argparse.ArgumentParser()

    def usage(args):
        parser.print_help()

    parser.set_defaults(func=usage)
    parser.add_argument(
        "--config", default="", help="path to custom config-file (default='')"
    )

    subparsers = parser.add_subparsers()

    # build
    build_parser = subparsers.add_parser("build")
    build_parser.set_defaults(func=build_compiled_annotation)
    build_parser.add_argument(
        "--gtf",
        default=None,
        required=True,
        help="path to the original annotation (e.g. gencodev38.gtf.gz)",
    )
    build_parser.add_argument(
        "--compiled",
        default=None,
        help="path to a directoy in which a compiled version of the GTF is stored",
    )
    build_parser.add_argument(
        "--tabular",
        default="",
        help="path to a cache of the tabular version of the relevant GTF features (optional)",
    )
    build_parser.add_argument(
        "--force-overwrite",
        default=False,
        action="store_true",
        help="re-compile GTF and overwrite the pre-existing compiled annotation",
    )

    # tag
    tag_parser = subparsers.add_parser("tag")
    tag_parser.set_defaults(func=main)
    tag_parser.add_argument(
        "--compiled",
        default=None,
        help="path to a directoy in which a compiled version of the GTF is stored",
    )
    tag_parser.add_argument(
        "--bam-in", default="/dev/stdin", help="path to the input BAM (default=stdin)"
    )
    tag_parser.add_argument(
        "--bam-out",
        default="/dev/stdout",
        help="path for the tagged BAM output (default=stdout)",
    )
    tag_parser.add_argument(
        "--bam-out-mode", default="b", help="mode of the output BAM file (default=b)"
    )
    tag_parser.add_argument(
        "--stats-out", default="", help="path for statistics output"
    )
    tag_parser.add_argument(
        "--threads-read",
        help="number of threads for reading bam_in (default=2)",
        type=int,
        default=2,
    )
    tag_parser.add_argument(
        "--threads-write",
        help="number of threads for writing bam_out (default=4)",
        type=int,
        default=4,
    )
    tag_parser.add_argument(
        "--threads-work",
        help="number of worker threads for actual trimming (default=8)",
        type=int,
        default=8,
    )

    # query
    query_parser = subparsers.add_parser("query")
    query_parser.set_defaults(func=query_regions)
    query_parser.add_argument(
        "--compiled",
        default=None,
        help="path to a directoy in which a compiled version of the GTF is stored",
    )
    query_parser.add_argument(
        "--gtf",
        default=None,
        help="path to the original annotation (e.g. gencodev38.gtf.gz)",
    )
    query_parser.add_argument("region", default=[], help="region to query", nargs="+")

    # gff
    gff_parser = subparsers.add_parser("gff")
    gff_parser.set_defaults(func=query_gff)
    gff_parser.add_argument(
        "--compiled",
        default=None,
        help="path to a directoy in which a compiled version of the GTF is stored",
    )
    gff_parser.add_argument(
        "--gff",
        default=None,
        help="path to the gff file you want queried",
    )

    args = parser.parse_args()
    return config.load(args.config, args=vars(args))


def cmdline():
    args = parse_args()
    util.setup_logging(args, "scbamtools.annotator.cmdline")
    return args.func(args)


if __name__ == "__main__":
    ret_code = cmdline()
    sys.exit(ret_code)

    # import cProfile
    # cProfile.run("cmdline()", "prof_stats")



#!/usr/bin/env python3
from scbamtools.contrib import __version__, __author__, __license__, __email__
import logging
import os
import sys
import numpy as np
import multiprocessing as mp
from collections import defaultdict
from time import time
import scbamtools.util as util
import mrfifo as mf

# TODO:
# * load params from YAML
# * save params to YAML in run-folder to document
# * refactor into multiple modules?


def FASTQ_src(src):
    from more_itertools import grouper

    for name, seq, _, qual in grouper(src, 4):
        yield name.rstrip()[1:], seq.rstrip(), qual.rstrip()


format_func_template = """
import re

def format_func(qname=None, r2_qname=None, r2_qual=None, r1=None, r2=None, R1=None, R2=None, **kw):
    #qparts = r2_qname.split(':N:0:')
    #if len(qparts) > 1:
    #    i5i7 = qparts[1].replace('+', '')
    #else:
    #    i5i7 = None

    cell = {cell}

    attrs = dict(
        cell=cell,
        raw=cell,
        UMI={UMI},
        seq={seq},
        qual={qual},
        fqid=r2_qname,
    )

    return attrs
"""


def safety_check_eval(s, danger="();."):
    chars = set(list(s))
    if chars & set(list(danger)):
        return False
    else:
        return True


def make_formatter_from_args(args):
    if not args.disable_safety:
        assert safety_check_eval(args.cell)
        assert safety_check_eval(args.UMI)
        assert safety_check_eval(args.seq)
        assert safety_check_eval(args.qual)

    # This trickery here compiles a python function that
    # embeds the user-specified code for cell, UMI
    # we replace globals() with an empty dictionary d,
    d = {}
    exec(
        format_func_template.format(
            cell=args.cell, UMI=args.UMI, seq=args.seq, qual=args.qual
        ),
        d,
    )
    # from said dict we can now retrieve the function object
    # and call it many times over
    func = d["format_func"]
    return func


def make_sam_record(
    fqid,
    seq,
    qual,
    flag=4,
    tags=[("CB", "{cell}"), ("MI", "{UMI}"), ("CR", "{raw}"), ("RG", "A")],
    **kw,
):
    tag_str = "\t".join([f"{tag}:Z:{tstr.format(**kw)}" for tag, tstr in tags])
    return f"{fqid}\t{flag}\t*\t0\t0\t*\t*\t0\t0\t{seq}\t{qual}\t{tag_str}\n"


def quality_trim(fq_src, min_qual=20, phred_base=33):
    for name, seq, qual in fq_src:
        end = len(seq)
        from cutadapt.qualtrim import quality_trim_index

        _, new_end = quality_trim_index(qual, 0, min_qual)
        # we use the cutadapt function here (which implements BWA's logic).
        # but only for the 3' end.
        n_trimmed = len(qual) - end

        # TODO: yield A3,T3 adapter-trimming tags
        if new_end != end:
            qual = qual[:new_end]
            seq = seq[:new_end]

        yield (name, seq, qual)


def render_to_sam(fq1, fq2, sam_out, args, **kwargs):
    logger = util.setup_logging(args, "scbamtools.bin.uBAM.worker")
    logger.debug(
        f"starting up with fq1={fq1}, fq2={fq2} sam_out={sam_out} and args={args}"
    )

    def iter_paired(fq1, fq2):
        fq_src1 = FASTQ_src(fq1)
        fq_src2 = FASTQ_src(fq2)
        if args.min_qual_trim:
            fq_src2 = quality_trim(
                fq_src2, min_qual=args.min_qual_trim, phred_base=args.phred_base
            )

        return zip(fq_src1, fq_src2)

    def iter_single(fq2):
        fq_src2 = FASTQ_src(fq2)
        if args.min_qual_trim:
            fq_src2 = quality_trim(
                fq_src2, min_qual=args.min_qual_trim, phred_base=args.phred_base
            )

        for fqid, seq, qual in fq_src2:
            yield (fqid, "NA", "NA"), (fqid, seq, qual)

    if fq1:
        ingress = iter_paired(fq1, fq2)
    else:
        ingress = iter_single(fq2)

    N = mf.util.CountDict()

    fmt = make_formatter_from_args(args)  # , **params

    for (fqid, r1, q1), (_, r2, q2) in ingress:
        N.count("total")
        attrs = fmt(r2_qname=fqid, r1=r1, r1_qual=q1, r2=r2, r2_qual=q2)
        sam_out.write(make_sam_record(flag=4, **attrs))

    return N
    # if args.paired_end:
    #             # if args.paired_end[0] == 'r':
    #             #     # rev_comp R1 and reverse qual1
    #             #     q1 = q1[::-1]
    #             #     r1 = util.rev_comp(r1)

    #             # if args.paired_end[1] == 'r':
    #             #     # rev_comp R2 and reverse qual2
    #             #     q2 = q2[::-1]
    #             #     r2 = util.rev_comp(r2)

    #             rec1 = fmt.make_bam_record(
    #                 qname=fqid,
    #                 r1="THISSHOULDNEVERBEREFERENCED",
    #                 r2=r1,
    #                 R1=r1,
    #                 R2=r2,
    #                 r2_qual=q1,
    #                 r2_qname=fqid,
    #                 flag=69,  # unmapped, paired, first in pair
    #             )
    #             rec2 = fmt.make_bam_record(
    #                 qname=fqid,
    #                 r1="THISSHOULDNEVERBEREFERENCED",
    #                 r2=r2,
    #                 R1=r1,
    #                 R2=r2,
    #                 r2_qual=q2,
    #                 r2_qname=fqid,
    #                 flag=133,  # unmapped, paired, second in pair
    #             )
    #             results.append(rec1)
    #             results.append(rec2)

    #         else:


def main(args):
    input_reads1, input_reads2, input_params = get_input_params(args)
    have_read1 = set([str(r1) != "None" for r1 in input_reads1]) == set([True])

    # queues for communication between processes
    w = (
        mf.Workflow("uBAM")
        # open reads2.fastq.gz
        .gz_reader(inputs=input_reads2, output=mf.FIFO("read2", "wb")).distribute(
            input=mf.FIFO("read2", "rt"),
            outputs=mf.FIFO("r2_{n}", "wt", n=args.parallel),
            chunk_size=args.chunk_size * 4,
        )
    )

    if have_read1:
        # open reads1.fastq.gz
        w.gz_reader(inputs=input_reads1, output=mf.FIFO("read1", "wb"))
        w.distribute(
            input=mf.FIFO("read1", "rt"),
            outputs=mf.FIFO("r1_{n}", "wt", n=args.parallel),
            chunk_size=args.chunk_size * 4,
        )
        # process in parallel workers
        w.workers(
            func=render_to_sam,
            fq1=mf.FIFO("r1_{n}", "rt"),
            fq2=mf.FIFO("r2_{n}", "rt"),
            sam_out=mf.FIFO("sam_{n}", "wt"),
            args=args,
            n=args.parallel,
        )
    else:
        # process in parallel workers
        w.workers(
            func=render_to_sam,
            fq1=None,
            fq2=mf.FIFO("r2_{n}", "rt"),
            sam_out=mf.FIFO("sam_{n}", "wt"),
            args=args,
            n=args.parallel,
        )

    # combine output streams
    w.collect(
        inputs=mf.FIFO("sam_{n}", "rt", n=args.parallel),
        chunk_size=args.chunk_size,
        custom_header=mf.util.make_SAM_header(
            prog_id="uBAM",
            prog_name="uBAM.py",
            prog_version=__version__,
            rg_name=args.sample,
        ),
        # output="/dev/stdout"
        # )
        output=mf.FIFO("sam_combined", "wt"),
    )
    # compress to BAM
    w.funnel(
        func=mf.parts.bam_writer,  # mf.parts.null_writer, #
        input=mf.FIFO("sam_combined", "rt"),
        output=args.bam_out,
        _manage_fifos=False,
        fmt="Sbh",
        threads=16,
    )
    return w.run()


def get_input_params(args):
    import pandas as pd
    import csv

    if str(args.matrix) != "None":
        df = pd.read_csv(args.matrix, sep=",", index_col=None, quoting=csv.QUOTE_NONE)
        if not "R1" in df.columns:
            df["R1"] = "None"

        R1 = df["R1"].to_list()
        R2 = df["R2"].to_list()
        # TODO extract other column values into a list of dictionaries (most importantly cell=...)
        # which can override how the formatter in the workers processes the raw reads
        if "cell" in df.columns:
            # params = [{'cell': f'"{c}"'} for c in df['cell']]
            params = [{"cell": c} for c in df["cell"]]
        else:
            params = [
                {},
            ] * len(R1)

    else:
        R1 = [
            args.read1,
        ]
        R2 = [
            args.read2,
        ]
        params = [
            {},
        ]

    return R1, R2, params


def parse_args():
    import scbamtools.util as util

    parser = util.make_minimal_parser(
        "uBAM.py",
        description="Convert raw reads1 and reads2 FASTQ into a single BAM file with cell barcode and UMI as BAM-tags",
    )
    parser.add_argument(
        "--matrix",
        default=None,
        help="sample_matrix.csv file desribing from where to get read1 and read2 (FASTQ format)",
    )
    parser.add_argument(
        "--read1",
        default=None,
        help="source from where to get read1 (FASTQ format)",
    )
    parser.add_argument(
        "--read2",
        default="/dev/stdin",
        help="source from where to get read2 (FASTQ format)",
        # required=True,
    )
    parser.add_argument("--cell", default="r1[8:20][::-1]")
    parser.add_argument("--UMI", default="r1[0:8]")
    parser.add_argument("--seq", default="r2")
    parser.add_argument("--qual", default="r2_qual")
    parser.add_argument("--disable-safety", default=False, type=bool)

    parser.add_argument(
        "--paired-end",
        default=None,
        choices=["fr", "ff", "rf", "rr"],
        help="read1 and read2 are paired end mates and store both in the BAM",
    )
    parser.add_argument(
        "--bam-out",
        default="/dev/stdout",
        help="output for unaligned BAM records (default=/dev/stdout) ",
    )
    parser.add_argument(
        "--save-stats",
        default="preprocessing_stats.txt",
        help="store statistics in this file",
    )
    parser.add_argument(
        "--save-cell-barcodes",
        default="",
        help="store (raw) cell barcode counts in this file. Numbers add up to number of total raw reads.",
    )
    parser.add_argument(
        "--fq-qual",
        default="E",
        help="phred qual for assigned barcode bases in FASTQ output (default='E')",
    )
    parser.add_argument(
        "--min-qual-trim",
        default=0,
        type=int,
        help="clip low quality-bases from a read2 3' end (e.g. pre-detected adapter)",
    )
    parser.add_argument(
        "--min-len",
        default=18,
        type=int,
        help="minimum read2 length to keep after quality trimming (default=18)",
    )
    parser.add_argument(
        "--phred-base",
        default=33,
        type=int,
        help="phred quality base (default=33)",
    )
    parser.add_argument(
        "--parallel", default=1, type=int, help="how many processes to spawn"
    )
    parser.add_argument(
        "--chunk-size",
        default=10,
        type=int,
        help="how many consecutive reads are assigned to the same worker (default=10)",
    )
    # SAM standard now supports CB=corrected cell barcode, CR=original cell barcode, and MI=molecule identifier/UMI
    parser.add_argument(
        "--bam-tags",
        default="CR:{cell},CB:{cell},MI:{UMI},RG:A",
        help="a template of comma-separated BAM tags to generate. Variables are replaced with extracted cell barcode, UMI etc.",
    )
    args = parser.parse_args()
    if args.parallel < 1:
        raise ValueError(f"--parallel {args.parallel} is invalid. Must be >= 1")

    return args


def cmdline():
    args = parse_args()
    util.setup_logging(args, name="scbamtools.bin.uBAM")

    if not args.read2:
        raise ValueError("bam output requires --read2 parameter")

    return main(args)


if __name__ == "__main__":
    res = cmdline()



from scbamtools.gene_model import *

genome = IndexedFasta("/data/rajewsky/genomes/GRCh38/GRCh38.fa")

w = 30
thresh = 0.6 * w
d = 50  # distance that is considered "close" to priming site
for chrom, size in genome.chrom_sizes.items():
    for pos in range(size - w):
        seq = genome.get_data(chrom, pos, pos + w, "+").upper()

        As = seq.count("A")
        Ts = seq.count("T")

        if As >= thresh and seq.startswith("AAA"):
            print(f"{chrom}\t{pos - d}\t{pos}\t{100*As/w:.0f}\t+\t{seq}")

        if Ts >= thresh and seq.endswith("TTT"):
            print(f"{chrom}\t{pos}\t{pos + d}\t{100*Ts/w:.0f}\t-\t{seq}")



from fixtures import *


def test_help():
    try:
        ubam("--help")
    except SystemExit:
        pass


def test_dropseq():
    ubam(
        "--read1",
        scbamtools_dir + "test_data/adap_test.R1.fastq",
        "--read2",
        scbamtools_dir + "test_data/adap_test.R2.fastq.gz",
        "--bam-out",
        "/dev/null",
        # "--flavor", "dropseq",
    )


def test_single():
    ubam(
        "--read2",
        scbamtools_dir + "test_data/adap_test.R2.fastq.gz",
        "--bam-out",
        "/dev/null",
        """--cell='"ACGTACGTACGTACGT"'""",
    )


def test_minqual():
    ubam(
        "--read2",
        scbamtools_dir + "test_data/adap_test.R2.fastq.gz",
        "--bam-out",
        "/dev/null",
        "--min-qual",
        "30",
        """--cell='"ACGTACGTACGTACGT"'""",
    )



#cython: boundscheck=False, wraparound=False, initializedcheck=False, overflowcheck=False, cdivision=True, language_level=3
##cython: boundscheck=True, wraparound=True, initializedcheck=True, overflowcheck=True, cdivision=False, language_level=3
#!python

__license__ = "MIT"
__authors__ = ["Marvin Jens"]
__email__ = "marvin.jens@charite.de"

from types cimport *
from types import *

import numpy as np
cimport numpy as np
cimport cython
cimport openmp

from libc.math cimport exp, log
from libc.stdlib cimport abort, malloc, free

cpdef inline CIGAR_to_blocks(str cigar, int pos):
    cdef list blocks = []
    cdef UINT32_t L = len(py_str)
    cdef UINT8_t x = 0
    cdef UINT8_t nt = 0
    cdef bytes py_byte_str = py_str.encode('ascii')
    cdef unsigned char *seq = py_byte_str
    
    cdef UINT64_t idx = 0
    for x in range(L):
        idx = (idx << 2) | letter_to_bits[seq[x]]

    return idx



import pytest
import os
import scbamtools.bctree as bt

@pytest.fixture(scope="session")
def test_compile(tmp_path_factory):
    tmp = tmp_path_factory.mktemp("bctree")
    scbamtools_path = os.path.dirname(__file__)
    compiled_path = tmp / "bctree.npy"

    bc = bt.compile([os.path.join(scbamtools_path, "../test_data/bc_test_data_0001.txt.gz")], dbname=compiled_path.as_posix())
    return compiled_path

@pytest.fixture(scope="session")
def test_load(test_compile):
    bc = bt.BCTree.from_disk(test_compile.as_posix())
    return bc

def test_batch(test_load):
    bc = test_load
    bc.check_contains_batch()

def test_hull(test_load):
    bc = test_load
    
    # edit distance 1 (aka "hull") generation benchmark
    for idx in bc.buf['idx'][:bc.n]:
        bt.bctree.hull(idx, 24)
    








from collections import namedtuple
import numpy as np
import re
from bisect import bisect_left, bisect_right

Gff = namedtuple(
    "gff",
    ["chrom", "source", "type", "start", "end", "score", "strand", "frame", "attr_str"],
)

Locus = namedtuple("Locus", ["name", "chrom", "start", "end", "strand", "type"])


def attr_to_dict(attr_str: str) -> dict:
    """
    Helper for reading the GTF. Parsers the attr_str found in the GTF
    attribute column (col 9). Returns a dictionary from the key, value
    pairs.
    """
    d = {}
    for match in re.finditer(r"(\w+) \"(\S+)\";", attr_str):
        key, value = match.groups()
        d[key] = value

    if not "gene_name" in d:
        d["gene_name"] = d.get("gene_id", d.get("transcript_id", "na"))

    return d


def GFF_importer(path):
    for line in open(path, "rt"):
        if line.startswith("#"):
            continue

        line = line.rstrip()
        if not line:
            continue

        cols = line.split("\t")
        cols[3] = int(cols[3])
        cols[4] = int(cols[4])
        # cols[-1] = attr_to_dict(cols[-1])

        yield Gff(*cols)


class ExonChain(object):
    """
    Implements linked blocks of genomic coordinates (with orientation).
    Features splicing from arbitrary tracks (including sequence) and intersection with
    simple start/end blocks on the genome. The latter returns 3 new ExonChains. As an
    example consider intersection with genomic CDS start and end-coordinates. This will
    return 5'UTR, CDS and 3'UTR as individual ExonChains.
    This class is deliberately kept as light-weight as possible. It has no means of
    generating names/annotation. All of that is handled by the "Transcript" subclass.
    """

    def __init__(self, chrom, strand, exon_starts, exon_ends, genome, name=""):
        self.genome = genome
        self.chrom = chrom
        self.strand = strand
        self.sense = strand  # deprecated
        self.exon_starts = np.array(sorted(exon_starts))
        self.exon_ends = np.array(sorted(exon_ends))
        self.name = name
        self._setup()

    def _setup(self):
        self.start = self.exon_starts[0]
        self.end = self.exon_ends[-1]

        self.exon_count = len(self.exon_starts)
        self.exon_bounds = np.array([self.exon_starts, self.exon_ends]).transpose()
        self.intron_bounds = []
        if self.exon_count > 1:
            self.intron_bounds = np.array(
                [self.exon_ends[:-1], self.exon_starts[1:]]
            ).transpose()

        self.exon_lengths = self.exon_ends - self.exon_starts
        self.exon_txstarts = np.array(
            [
                0,
            ]
            + list(self.exon_lengths.cumsum())
        )

        self.spliced_length = max(self.exon_txstarts[-1], 0)
        self.unspliced_length = self.end - self.start

        if self.strand == "-":
            self.dir = -1
            self.ofs = self.spliced_length
        else:
            self.dir = +1
            self.ofs = 0

        self.tx_start, self.tx_end = [self.start, self.end][:: self.dir]
        if not self.name:
            # if the chain gets altered subsequently, don't loose the name
            self.name = "ExonChain_%s:%d-%d%s__%s__%s" % (
                self.chrom,
                self.start,
                self.end,
                self.strand,
                ",".join([str(s) for s in self.exon_starts]),
                ",".join([str(e) for e in self.exon_ends]),
            )

    def upstream(self, L=100):
        if self.strand == "+":
            exon_starts = [self.start - L]
            exon_ends = [self.start]
        else:
            exon_starts = [self.end]
            exon_ends = [self.end + L]

        return ExonChain(
            self.chrom,
            self.strand,
            exon_starts,
            exon_ends,
            self.genome,
            f"{self.name}_upstream_{L}",
        )

    def downstream(self, L=100):
        if self.strand == "+":
            exon_starts = [self.end]
            exon_ends = [self.end + L]
        else:
            exon_starts = [self.start - L]
            exon_ends = [self.start]

        return ExonChain(
            self.chrom,
            self.strand,
            exon_starts,
            exon_ends,
            self.genome,
            f"{self.name}_downstream_{L}",
        )

    @property
    def intron_chain(self):
        if self.exon_count > 1:
            return ExonChain(
                self.chrom,
                self.strand,
                self.exon_ends[:-1],
                self.exon_starts[1:],
                genome=self.genome,
            )

    def map_to_spliced(self, pos, truncate=False):
        if self.strand == "+":
            n = min(bisect_right(self.exon_starts, pos), self.exon_count) - 1
            if not (self.exon_starts[n] <= pos <= self.exon_ends[n]):
                if not truncate:
                    raise ValueError("%d does not lie within any exon bounds" % pos)
                else:
                    pos = self.exon_ends[n]

            return self.ofs + self.dir * (
                pos - self.exon_starts[n] + self.exon_txstarts[n]
            )
        else:
            n = min(max(0, bisect_left(self.exon_ends, pos)), self.exon_count - 1)
            if not (self.exon_starts[n] <= pos <= self.exon_ends[n]):
                if not truncate:
                    raise ValueError("%d does not lie within any exon bounds" % pos)
                else:
                    pos = self.exon_starts[n]

            return (
                self.ofs
                + self.dir * (pos - self.exon_starts[n] + self.exon_txstarts[n])
                - 1
            )

    def map_block_from_spliced(self, start, end):
        x, y = self.map_from_spliced(start), self.map_from_spliced(end)
        if y > x:
            # semantics change here because start is included, end is excluded
            # in C indexing.
            return x, y
        else:
            return y, x
        # return min(x,y),max(x,y)

    def map_block_to_spliced(self, start, end, truncate=False):
        x, y = self.map_to_spliced(start, truncate), self.map_to_spliced(end, truncate)
        if y > x:
            # semantics change here because start is included, end is excluded
            # in C indexing.
            return x, y
        else:
            return y, x

    def map_to_exon(self, tx_pos):
        pos = self.ofs + self.dir * tx_pos
        n = max(0, bisect_left(self.exon_txstarts, pos) - 1)
        return n

    def map_from_spliced(self, pos):
        pos = self.ofs + self.dir * pos
        if self.strand == "+":
            n = min(bisect_right(self.exon_txstarts, pos), self.exon_count) - 1
            if n == self.exon_count or n < 0:
                raise ValueError("{0} out of transcript bounds".format(pos))

            return self.exon_starts[n] + pos - self.exon_txstarts[n]
        else:
            n = max(0, bisect_left(self.exon_txstarts, pos) - 1)
            if n == self.exon_count or n < 0:
                raise ValueError("{0} out of transcript bounds".format(pos))

            return self.exon_starts[n] + pos - self.exon_txstarts[n] - 1

    def splice(self, track, join=lambda l: "".join(l), get="get_oriented", **kwargs):
        get_func = getattr(track, get)
        return join(
            [
                get_func(self.chrom, start, end, self.strand, **kwargs)
                for start, end in self.exon_bounds
            ][:: self.dir]
        )

    @property
    def unspliced_sequence(self):
        return self.genome.get_oriented(self.chrom, self.start, self.end, self.strand)

    @property
    def spliced_sequence(self):
        return self.splice(self.genome)

    @property
    def splice_sites(self):
        """returns tuples of (5'ss, 3'ss)"""
        for bounds in self.intron_bounds[:: self.dir]:
            yield bounds[:: self.dir]

    @property
    def exons(self):
        for i, (start, end) in enumerate(self.exon_bounds[:: self.dir]):
            yield Locus(
                "%s/exon.%02d" % (self.name, i + 1),
                self.chrom,
                self.strand,
                start,
                end,
                "exon",
            )

    @property
    def introns(self):
        for i, (start, end) in enumerate(self.intron_bounds[:: self.dir]):
            yield Locus(
                "%s/intron.%02d" % (self.name, i + 1),
                self.chrom,
                self.strand,
                start,
                end,
                "intron",
            )

    @property
    def introns_as_chains(self):
        for i, (start, end) in enumerate(self.intron_bounds[:: self.dir]):
            yield ExonChain(
                self.chrom,
                self.strand,
                [
                    start,
                ],
                [
                    end,
                ],
                name="%s/intron.%02d" % (self.name, i + 1),
                genome=self.genome,
            )

    def cut(self, start, end, expand=False):
        names = ["before", "cut", "after"]
        return self.intersect(names, start, end, expand=expand)

    def intersect(self, names, start, end, expand=False):
        # TODO: Clean this up
        # print "INTERSECT",names,start,end
        start, end = min(start, end), max(start, end)
        if not expand:
            start = max(start, self.start)
            end = min(end, self.end)

        first = bisect_right(self.exon_starts, start) - 1
        last = bisect_right(self.exon_starts, end) - 1

        f2 = bisect_right(self.exon_ends, start)
        l2 = bisect_right(self.exon_ends, end)

        # print first,f2
        # print last,l2

        before_starts = list(self.exon_starts[: f2 + 1])
        before_ends = list(self.exon_ends[: f2 + 1])

        if not before_starts and not before_ends:
            before_starts = [0]
            before_ends = [0]

        chain_starts = list(self.exon_starts[f2 : last + 1])
        chain_ends = list(self.exon_ends[f2 : last + 1])

        if not chain_starts and not chain_ends:
            chain_starts = [0]
            chain_ends = [0]

        # if not chain_starts:
        # print first,last,start,end,self.tx_start,self.tx_end

        after_starts = list(self.exon_starts[last:])
        after_ends = list(self.exon_ends[last:])

        if not after_starts and not after_ends:
            after_starts = [0]
            after_ends = [0]

        if first == f2:
            chain_starts[0] = max(start, chain_starts[0])
            before_ends[-1] = min(start, before_ends[-1])
        else:
            # print("start falls between exon bounds?")
            # truncate chain, breakpoint between two exons
            before_ends[-1] = before_starts[-1]
            if expand:
                # expand the exon-bound to incorporate the start
                chain_starts[0] = start

        if last == l2:
            chain_ends[-1] = min(end, chain_ends[-1])
            after_starts[0] = max(end, after_starts[0])
        else:
            # print("end falls between exon bounds?")
            # truncate chain, breakpoint between two exons
            after_starts[0] = after_ends[0]
            if expand:
                # expand the exon-bound to incorporate the end
                chain_ends[-1] = end

        def remove_empty(starts, ends):
            keep_starts = []
            keep_ends = []
            for s, e in zip(starts, ends):
                if s != e:
                    keep_starts.append(s)
                    keep_ends.append(e)

            return keep_starts, keep_ends

        before_starts, before_ends = remove_empty(before_starts, before_ends)
        chain_starts, chain_ends = remove_empty(chain_starts, chain_ends)
        after_starts, after_ends = remove_empty(after_starts, after_ends)

        if chain_starts:
            chain = ExonChain(
                self.chrom, self.strand, chain_starts, chain_ends, genome=self.genome
            )
        else:
            chain = self.EmptyChain()

        if before_starts:
            before = ExonChain(
                self.chrom, self.strand, before_starts, before_ends, genome=self.genome
            )
        else:
            before = self.EmptyChain()

        if after_starts:
            after = ExonChain(
                self.chrom, self.strand, after_starts, after_ends, genome=self.genome
            )
        else:
            after = self.EmptyChain()

        if self.strand == "+":
            return before, chain, after
        else:
            return after, chain, before

    def cut(self, start, end, expand=False):
        before, chain, after = self.intersect(
            ["before", "cut", "after"], start, end, expand=expand
        )
        return chain

    @property
    def key(self):
        return (self.chrom, self.strand, tuple(self.exon_starts), tuple(self.exon_ends))

    @property
    def key_str(self):
        return "{self.chrom}:{es}-{ee}{self.strand}".format(
            self=self,
            es=",".join([str(x) for x in self.exon_starts]),
            ee=",".join([str(x) for x in self.exon_ends]),
        )

    def __add__(self, other):
        """
        concatenates two non-overlapping exon-chains, returning a new ExonChain object
        """

        # ensure A precedes B in chromosome
        if self.start < other.start:
            A, B = self, other
        else:
            A, B = other, self

        if not A:
            return B
        if not B:
            return A
        # print "# concatenating exon chains",A,B
        assert A.chrom == B.chrom
        assert A.strand == B.strand
        assert A.genome == B.genome
        assert A.end <= B.start  # support only non-overlapping chains for now!

        # these must now stay in order and can not overlap!
        exon_starts = np.concatenate((A.exon_starts, B.exon_starts))
        exon_ends = np.concatenate((A.exon_ends, B.exon_ends))

        return ExonChain(A.chrom, A.strand, exon_starts, exon_ends, genome=A.genome)

    # add some python magic to make things smooth
    def __str__(self):
        exonlist = ",".join(map(str, self.exon_bounds))
        return "{self.chrom}:{self.start}-{self.end}{self.strand} spliced_length={self.spliced_length} exons: {exonlist}".format(
            self=self, exonlist=exonlist
        )

    def bed12_format(self, color="255,0,0"):
        block_sizes = [str(e - s) for s, e in self.exon_bounds]
        block_starts = [str(s - self.start) for s in self.exon_starts]

        if getattr(self, "CDS", None):
            cds_start = self.CDS.start
            cds_end = self.CDS.end
        else:
            cds_start = self.end
            cds_end = self.end

        cols = [
            self.chrom,
            self.start,
            self.end,
            self.name,
            getattr(self, "score", 0),
            self.strand,
            cds_start,
            cds_end,
            color,
            self.exon_count,
            ",".join(block_sizes),
            ",".join(block_starts),
        ]
        return "\t".join([str(c) for c in cols])

    def ucsc_format(self):
        exonstarts = ",".join([str(s) for s in self.exon_starts])
        exonends = ",".join([str(e) for e in self.exon_ends])
        # exonframes = ",".join([str(f) for f in self.exon_frames])
        # TODO: fix exon-frames
        exonframes = ",".join(["-1" for e in self.exon_ends])

        if getattr(self, "CDS", None):
            cds_start = self.CDS.start
            cds_end = self.CDS.end
        else:
            cds_start = self.end
            cds_end = self.end

        out = (
            -1,
            self.name,
            self.chrom,
            self.strand,
            self.start,
            self.end,
            cds_start,
            cds_end,
            self.exon_count,
            exonstarts,
            exonends,
            getattr(self, "score", 0),
            getattr(self, "gene_id", self.name),
            "unk",
            "unk",
            exonframes,
        )
        return "\t".join([str(o) for o in out])

    def __len__(self):
        """
        zero-length ExonChains will be False in truth value testing.
        so stuff like: "if tx.UTR5" can work.
        """
        return self.spliced_length

    def __hasitem__(self, pos):
        """check if the coordinate falls into an exon"""
        try:
            x = self.map_to_spliced(pos)
        except ValueError:
            return False
        else:
            return True

    def EmptyChain(self):
        return ExonChain(self.chrom, self.strand, [0], [0], genome=self.genome)


class Transcript(ExonChain):
    def __init__(
        self,
        name,
        chrom,
        sense,
        exon_starts,
        exon_ends,
        cds,
        score=0,
        genome=None,
        description={},
        **kwargs,
    ):
        # Initialize underlying ExonChain
        super(Transcript, self).__init__(
            chrom, sense, exon_starts, exon_ends, genome=genome
        )
        self.score = score
        self.category = "transcript/body"
        self.transcript_id = kwargs.get("transcript_id", name)
        self.gene_id = kwargs.get("gene_id", name)
        self.gene_name = kwargs.get("gene_name", name)
        self.gene_type = kwargs.get("gene_type", "unknown")

        self.name = name
        self.description = description

        cds_start, cds_end = cds
        if cds_start == cds_end:
            self.UTR5 = None
            self.UTR3 = None
            self.CDS = None
            self.coding = False
            self.tx_class = description.get("tx_class", "ncRNA")
        else:
            self.UTR5, self.CDS, self.UTR3 = self.intersect(
                ["UTR5", "CDS", "UTR3"], cds_start, cds_end
            )
            if not (self.CDS.start == cds_start and self.CDS.end == cds_end):
                print(self.gene_name, self.name)
                print(self.CDS)
                print(cds_start, cds_end)
                1 / 0

            self.coding = True
            self.tx_class = description.get("tx_class", "coding")
            if self.UTR5:
                self.UTR5.name = "%s/UTR5" % self.name
                self.UTR5.gene_id = self.gene_id
                self.UTR5.category = "transcript/UTR5"

            if self.CDS:
                self.CDS.name = "%s/CDS" % self.name
                self.CDS.gene_id = self.gene_id
                self.UTR5.category = "transcript/CDS"

            if self.UTR3:
                self.UTR3.name = "%s/UTR3" % self.name
                self.UTR3.gene_id = self.gene_id
                self.UTR5.category = "transcript/UTR3"

            if self.strand == "+":
                self.CDS.start_codon = cds_start
                self.CDS.stop_codon = cds_end
            else:
                self.CDS.start_codon = cds_end
                self.CDS.stop_codon = cds_start

    @classmethod
    def from_GTF(cls, fname="", ORF_thresh=20, genome=None):
        shared_attributes = {}
        exon_starts = []
        exon_ends = []
        chrom = "NN"
        sense = "NN"
        current_tx = "NN"

        n = 0
        cds_min = np.inf
        cds_max = -np.inf
        ignore_records = set(["gene", "start_codon", "stop_codon", "UTR", "transcript"])
        for i, gff in enumerate(GFF_importer(fname)):
            # sys.stderr.write(".")
            if gff.type in ignore_records:
                continue

            attrs = attr_to_dict(gff.attr_str)
            tx_id = attrs.get("transcript_id", "unknown_tx_{}".format(n))
            if tx_id != current_tx:
                if exon_starts:
                    if cds_min is np.inf:
                        cds = (exon_starts[0], exon_starts[0])
                    else:
                        cds = (cds_min, cds_max)

                    yield cls(
                        current_tx,
                        chrom,
                        sense,
                        exon_starts,
                        exon_ends,
                        cds,
                        genome=genome,
                        **shared_attributes,
                    )

                exon_starts = []
                exon_ends = []
                shared_attributes = {}
                cds_min = np.inf
                cds_max = -np.inf
                n += 1

                current_tx = tx_id
                chrom = gff.chrom
                sense = gff.strand

            if gff.type == "CDS":
                cds_min = min(gff.start - 1, cds_min)
                cds_max = max(gff.end, cds_max)
                # print("encountered CDS record. current CDS", cds_min, cds_max)
                continue

            if gff.type != "exon":
                continue

            shared_attributes.update(attrs)

            exon_starts.append(gff.start - 1)
            exon_ends.append(gff.end)

        if exon_starts:
            if cds_min is np.inf:
                cds = (exon_starts[0], exon_starts[0])
            else:
                cds = (cds_min, cds_max)

            yield cls(
                tx_id,
                chrom,
                sense,
                exon_starts,
                exon_ends,
                cds,
                description=shared_attributes,
                genome=genome,
                **shared_attributes,
            )

    @classmethod
    def from_bed12(cls, line, genome=None):
        (
            chrom,
            start,
            end,
            name,
            score,
            strand,
            cds_start,
            cds_end,
            color,
            exon_count,
            block_sizes,
            block_starts,
        ) = line.rstrip().split("\t")

        start = int(start)
        end = int(end)

        exon_starts = np.array(block_starts.split(","), dtype=int) + start
        exon_ends = np.array(block_sizes.split(","), dtype=int) + exon_starts

        chain = cls(
            name,
            chrom,
            strand,
            exon_starts,
            exon_ends,
            (int(cds_start), int(cds_end)),
            score=np.nan if score == "." else float(score),
            genome=genome,
            description=dict(color=color, dtype=np.uint8),
        )

        return chain

    # UCSC table format output
    def __str__(self):
        return self.ucsc_format()


def _maybe_file(src):
    if hasattr(src, "read"):
        return src
    else:
        return open(src)


def from_bed6(src, genome=None):
    import numpy as np

    for line in _maybe_file(src):
        chrom, start, end, name, score, strand = line.rstrip().split("\t")
        chain = ExonChain(
            chrom,
            strand,
            [
                int(start),
            ],
            [
                int(end),
            ],
            genome=genome,
        )
        chain.name = name
        if score == ".":
            chain.score = np.nan
        else:
            chain.score = float(score)
        yield chain


def from_bed12(src, genome=None, filter=None):
    import numpy as np

    for line in _maybe_file(src):
        if not (filter is None) and not (filter in line):
            continue

        chain = Transcript.from_bed12(line, genome=genome)
        yield chain


class IndexedFasta(object):
    def __init__(self, fname, split_chrom="", **kwargs):
        import logging
        import mmap
        import os

        self.logger = logging.getLogger("scbamtools.gene_model.IndexedFasta")

        self.fname = fname
        self.chrom_stats = {}
        self.chrom_sizes = {}
        self.split_chrom = split_chrom

        f = open(fname)
        idx_file = f
        self._f = mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ)

        # try to load index
        ipath = fname + ".byo_index"
        if os.access(ipath, os.R_OK):
            self.load_index(ipath)
        else:
            self.index(idx_file)
            self.store_index(ipath)

    def index(self, idx_file):
        self.logger.debug(
            "index('{self.fname}') split_chrom={self.split_chrom}".format(**locals())
        )

        ofs = 0
        chrom = "undef"
        chrom_ofs = 0
        size = 0
        nl_char = 0

        for line in idx_file:
            ofs += len(line)
            if line.startswith(">"):
                # store size of previous sequence
                if size:
                    self.chrom_stats[chrom].append(size)

                chrom = line[1:].split()[0].strip()
                if self.split_chrom:
                    # use this to strip garbage from chrom/contig name like chr1:new
                    # ->split_chrom=':' -> chrom=chr1
                    chrom = chrom.split(self.split_chrom)[0]
                chrom_ofs = ofs
            else:
                if not chrom in self.chrom_stats:
                    # this is the first line of the new chrom
                    size = 0
                    lline = len(line)
                    ldata = len(line.strip())
                    nl_char = lline - ldata
                    self.chrom_stats[chrom] = [chrom_ofs, ldata, nl_char, line[ldata:]]
                size += len(line.strip())

        # store size of previous sequence
        if size:
            self.chrom_stats[chrom].append(size)
        idx_file.flush()
        idx_file.seek(0)

    def store_index(self, ipath):
        import os

        self.logger.info("store_index('%s')" % ipath)

        # write to tmp-file first and in the end rename in order to have this atomic
        # otherwise parallel building of the same index may screw it up.

        import tempfile

        tmp = tempfile.NamedTemporaryFile(
            mode="w", dir=os.path.dirname(ipath), delete=False
        )
        for chrom in sorted(self.chrom_stats.keys()):
            ofs, ldata, skip, skipchar, size = self.chrom_stats[chrom]
            tmp.write(
                "%s\t%d\t%d\t%d\t%r\t%d\n" % (chrom, ofs, ldata, skip, skipchar, size)
            )

        # make sure everything is on disk
        os.fsync(tmp)
        tmp.close()

        # make it accessible to everyone
        import stat

        os.chmod(tmp.name, stat.S_IROTH | stat.S_IRGRP | stat.S_IRUSR)

        # this is atomic on POSIX as we have created tmp in the same directory,
        # therefore same filesystem
        os.rename(tmp.name, ipath)

    def load_index(self, ipath):
        self.logger.info("load_index('%s')" % ipath)
        self.chrom_stats = {}
        for line in open(ipath):
            chrom, ofs, ldata, skip, skipchar, size = line.rstrip().split("\t")
            self.chrom_stats[chrom] = (
                int(ofs),
                int(ldata),
                int(skip),
                skipchar[1:-1].encode("ascii").decode("unicode_escape"),
                int(size),
            )
            self.chrom_sizes[chrom] = int(size)

    def get_data(self, chrom, start, end, sense):
        from scbamtools.util import rev_comp

        if not self.chrom_stats:
            self.index()

        ofs, ldata, skip, skip_char, size = self.chrom_stats[chrom]
        # print("ldata",ldata)
        # print("chromstats",self.chrom_stats[chrom])
        pad_start = 0
        pad_end = 0
        if start < 0:
            pad_start = -start
            start = 0

        if end > size:
            pad_end = end - size
            end = size

        l_start = int(start / ldata)
        l_end = int(end / ldata)
        # print "lines",l_start,l_end
        ofs_start = l_start * skip + start + ofs
        ofs_end = l_end * skip + end + ofs
        # print("ofs",ofs_start,ofs_end,ofs_end - ofs_start)
        # print(type(skip_char))

        s = self._f[ofs_start:ofs_end].decode("ascii").replace(skip_char, "")
        if pad_start or pad_end:
            s = "N" * pad_start + s + "N" * pad_end

        if sense == "-":
            s = rev_comp(s)
        return s


if __name__ == "__main__":
    genome = IndexedFasta("/data/rajewsky/genomes/GRCh38/GRCh38.fa")
    for tx in Transcript.from_GTF("/dev/stdin", genome=genome):
        print(tx.upstream(100).bed12_format())
        print(tx.downstream(100).bed12_format())



import numpy as np
import numpy
import pyximport

pyximport.install(setup_args={"include_dirs": numpy.get_include()}, reload_support=True)
import scbamtools.cython.bctree as bctree
from time import time
import logging

LEAF = 0xFFFFFFFE


class BCTree:
    logger = logging.getLogger("spacemake.BCTree")
    dtype = [
        ("idx", np.uint64),
        ("left", np.uint32),
        ("right", np.uint32),
        ("x", np.uint16),
        ("y", np.uint16),
        ("tile", np.uint16),
    ]

    def __init__(self, N=int(2e9)):
        self.N = N
        self.n = 0
        self.buf = np.zeros(N, dtype=self.dtype)
        self.root = None

    def load_tile_data(self, fnames):
        import gzip

        t0 = time()
        n = 0
        for fname in fnames:
            tile = int(fname.split(".txt")[0].split("_")[-1])
            if fname.endswith(".gz"):
                src = gzip.open(fname, "rt")
            else:
                src = open(fname)

            n_read = bctree.read_txt_to_buf(src, self.buf[n:], tile=tile)
            n += n_read

        self.n = n
        t1 = time()
        self.logger.info(
            f"{1000 * (t1-t0):.2f} msec loading and converting {self.n} barcode records"
        )

        # truncate array
        self.N = self.n
        self.buf = self.buf[:n]

    def sort(self):
        # TODO: replace with faster, parallel implementation (e.g. parallel numpy)
        try:
            import pnumpy as pn

            pn.thread_setworkers(16)
        except ImportError:
            self.logger.warning(
                "Can not import pnumpy (parallel numpy). Sorting will be slow."
            )

        t0 = time()

        ## in-place sorting is slow
        # self.buf[:self.n].sort()

        # faster to sort on the idx integers only and then re-order
        I = self.buf["idx"][: self.n].argsort()
        self.buf = self.buf[I]

        t1 = time()
        dt = t1 - t0
        self.logger.info(f"{1000 * dt:.2f} msec sorting {self.n} barcode records")

    def build_BST(self):
        t0 = time()

        # build-tree
        self.root = bctree.build_tree(self.buf, 0, self.n)

        t1 = time()
        dt = t1 - t0
        self.logger.info(f"{1000 * dt:.2f} msec building {self.n} barcode BST")

    def check_contains_batch(self):
        assert self.root is not None
        t0 = time()

        # checking in batch
        bctree.check_contains_batch(self.buf, self.root, self.buf["idx"][: self.n])

        t1 = time()
        dt = t1 - t0
        self.logger.info(
            f"{1000 * dt:.2f} msec batch checking for match of {self.n} existing barcodes ({0.001 * self.n/dt:.2f} kBC/sec)"
        )

    def save_to_disk(self, filename):
        np.save(filename, self.buf)

    @classmethod
    def from_disk(cls, filename):
        buf = np.lib.format.open_memmap(filename, dtype=cls.dtype, mode="r")
        bc = cls(N=len(buf))
        bc.n = len(buf)
        bc.buf = buf
        bc.root = bc.n >> 1

        bc.logger.info(
            f"loaded BCTree data for {bc.n} barcodes from file '{filename}'."
        )
        return bc


def compile(fnames, dbname="bctree.npy"):
    bc = BCTree()
    bc.load_tile_data(fnames)
    bc.sort()
    bc.build_BST()

    bc.save_to_disk(dbname)
    return bc


# if __name__ == "__main__":
# logging.basicConfig(level=logging.INFO)

## 2.4M smallest example. Should be very fast
# bc = compile(["fc_010_L3_tile_2267.txt"])

## 8M barcodes
# bc = compile(["fc_010_L3_tile_2267.txt", "fc_010_L3_tile_2268.txt", "fc_010_L3_tile_2269.txt"], 'bctree.npy')

## 216M barcodes + testing gz input -> takes ~4.5 GB on disk
# from glob import glob
# fnames = glob("/data/rajewsky/projects/slide_seq/puck_data/seq_scope_inhouse_fc_sts_nova5_FC_010_230109/fc_010_L3_tile_22*.gz")
# print(fnames)
# bc = compile(fnames)

# bc.check_contains_batch()

# bc = BCTree.from_disk('bctree.npy')
# bc.check_contains_batch()

# t5 = time()
# # edit distance 1 (aka "hull") generation benchmark
# for idx in bc.buf['idx'][:bc.n]:
#     bctree.hull(idx, 24)

# t6 = time()
# dt = t6 - t5
# print(f"{1000 * dt:.2f} msec hull generation for {bc.n} existing barcodes ({0.001 * bc.n/dt:.2f} kBC/sec)")


# t5 = time()
# # edit distance 1 (aka "hull") generation benchmark
# for idx in bc.buf['idx'][:bc.n]:
#     bctree.hull(idx, 24)

# t6 = time()
# dt = t6 - t5
# print(f"{1000 * dt:.2f} msec hull generation for {bc.n} existing barcodes ({0.001 * bc.n/dt:.2f} kBC/sec)")



import logging
import os
import sys

default_log_level = "INFO"


def ensure_path(path):
    dirname = os.path.dirname(path)
    if dirname:
        os.makedirs(dirname, exist_ok=True)
    return path


def setup_logging(
    args,
    name="scbamtools.main",
    log_file="",
    FORMAT="%(asctime)-20s\t{sample:30s}\t%(name)-50s\t%(levelname)s\t%(message)s",
):
    sample = getattr(args, "sample", "na")
    import setproctitle

    if name != "scbamtools.main":
        setproctitle.setproctitle(f"{name} {sample}")

    FORMAT = FORMAT.format(sample=sample)

    log_level = getattr(args, "log_level", "INFO")
    lvl = getattr(logging, log_level)
    logging.basicConfig(level=lvl, format=FORMAT)
    root = logging.getLogger("spacemake")
    root.setLevel(lvl)

    log_file = getattr(args, "log_file", log_file)
    if log_file:
        fh = logging.FileHandler(filename=ensure_path(log_file), mode="a")
        fh.setFormatter(logging.Formatter(FORMAT))
        root.debug(f"adding log-file handler '{log_file}'")
        root.addHandler(fh)

    if hasattr(args, "debug"):
        # cmdline requested debug output for specific domains (comma-separated)
        for logger_name in args.debug.split(","):
            if logger_name:
                root.info(f"setting domain {logger_name} to DEBUG")
                logging.getLogger(logger_name.replace("root", "")).setLevel(
                    logging.DEBUG
                )

    logger = logging.getLogger(name)
    logger.debug("started logging")
    for k, v in sorted(vars(args).items()):
        logger.debug(f"cmdline arg\t{k}={v}")

    return logger


def make_minimal_parser(prog="", usage="", main=None, **kw):
    import argparse

    parser = argparse.ArgumentParser(prog=prog, usage=usage, **kw)
    parser.add_argument(
        "--log-file",
        default="",  # f"{prog}.log",
        help=f"place log entries in this file (default=off)",
    )
    parser.add_argument(
        "--log-level",
        default=default_log_level,
        help=f"change threshold of python logging facility (default={default_log_level})",
    )
    parser.add_argument(
        "--debug",
        default="",
        help=f"comma-separated list of logging-domains for which you want DEBUG output",
    )
    parser.add_argument(
        "--sample", default="sample_NA", help="sample_id (where applicable)"
    )
    parser.add_argument(
        "--version",
        default=False,
        action="store_true",
        help="output version information and exit",
    )

    def default_logic(args):
        if args.version:
            from scbamtools.contrib import __version__, __author__, __license__

            print(f"scbamtools {prog} version {__version__}")
            print(f"by {__author__}, released under {__license__} license")

        else:
            import scbamtools.config as config

            if hasattr(args, "config"):
                args = config.load(args.config, args=vars(args))

            setup_logging(args)

            return main(args)

    parser.set_defaults(func=default_logic)

    return parser


def update_header(input, output, progname="scbamtools", cmdline=" ".join(sys.argv)):
    from collections import defaultdict
    from scbamtools.contrib import __version__

    id_counter = defaultdict(int)
    pp = ""
    for header_line in input:
        if header_line.startswith("@PG"):
            parts = header_line.split("\t")
            for part in parts[1:]:
                k, v = part.split(":", maxsplit=1)
                if k == "ID":
                    pp = v
                    id_counter[v] += 1

        output.write(header_line)

    pg_id = progname.split(".")[0]
    if id_counter[pg_id] > 0:
        pg_id += f".{id_counter[pg_id]}"

    output.write(
        f"@PG\tID:{pg_id}\tPN:{progname}\tPP:{pp}\tVN:{__version__}\tCL:{cmdline}\n"
    )


def generate_kmers(k, nts="ACGT"):
    if k == 0:
        yield ""
    elif k > 0:
        for x in nts:
            for mer in generate_kmers(k - 1, nts=nts):
                yield x + mer


COMPLEMENT = {
    "a": "t",
    "t": "a",
    "c": "g",
    "g": "c",
    "k": "m",
    "m": "k",
    "r": "y",
    "y": "r",
    "s": "s",
    "w": "w",
    "b": "v",
    "v": "b",
    "h": "d",
    "d": "h",
    "n": "n",
    "A": "T",
    "T": "A",
    "C": "G",
    "G": "C",
    "K": "M",
    "M": "K",
    "R": "Y",
    "Y": "R",
    "S": "S",
    "W": "W",
    "B": "V",
    "V": "B",
    "H": "D",
    "D": "H",
    "N": "N",
    "-": "-",
    "=": "=",
    "+": "+",
}


def complement(s):
    return "".join([COMPLEMENT[x] for x in s])


def rev_comp(seq):
    return complement(seq)[::-1]


def is_header(line):
    return line.startswith("@")



import pytest
import os
from scbamtools.annotation import GenomeAnnotation, CompiledClassifier

@pytest.fixture(scope="session")
def test_compile(tmp_path_factory):
    tmp = tmp_path_factory.mktemp("simple")
    scbamtools_path = os.path.dirname(__file__)
    ga = GenomeAnnotation.from_GTF(
        os.path.join(scbamtools_path, "../test_data/simple_annotation.gtf"),
        df_cache=(tmp / "simple.tsv").as_posix()
    )
    compiled_path = (tmp / "compiled/").as_posix()
    _ = ga.compile(compiled_path)
    
    return tmp / "compiled"

def test_recompile(test_compile):
    assert CompiledClassifier.files_exist(test_compile)
    ga = GenomeAnnotation.from_compiled_index(test_compile)
    try:
        ga.compile()
    except ValueError:
        pass
    else:
        raise ValueError("attempt to re-compile already compiled annotation should have failed!")
    
    ga2 = GenomeAnnotation.from_uncompiled_df((test_compile / "../simple.tsv").as_posix()).compile(test_compile.as_posix())
    assert (ga2.classifier.classifications == ga.classifier.classifications).all()


def test_regions(test_compile):
    """
    chr1    test    transcript      1       10000   .       +       .       gene_name "A"; transcript_type "protein_coding"; transcript_id "A.1";
    chr1    test    exon    1       150     .       +       .       gene_name "A"; transcript_type "protein_coding"; transcript_id "A.1";
    chr1    test    exon    9000    10000   .       +       .       gene_name "A"; transcript_type "protein_coding"; transcript_id "A.1";
    chr1    test    CDS     75      150     .       +       .       gene_name "A"; transcript_type "protein_coding"; transcript_id "A.1";
    chr1    test    CDS     9000    9050    .       +       .       gene_name "A"; transcript_type "protein_coding"; transcript_id "A.1";

    chr2    test    transcript      1       10000   .       +       .       gene_name "B"; transcript_type "protein_coding"; transcript_id "B.1";
    chr2    test    exon    1       150     .       +       .       gene_name "B"; transcript_type "protein_coding"; transcript_id "B.1";
    chr2    test    exon    9000    10000   .       +       .       gene_name "B"; transcript_type "protein_coding"; transcript_id "B.1";
    chr2    test    transcript      1       10000   .       +       .       gene_name "B"; transcript_type "protein_coding"; transcript_id "B.2";
    chr2    test    exon    1       150     .       +       .       gene_name "B"; transcript_type "protein_coding"; transcript_id "B.2";
    chr2    test    exon    500     600     .       +       .       gene_name "B"; transcript_type "protein_coding"; transcript_id "B.2";
    chr2    test    exon    9000    10000   .       +       .       gene_name "B"; transcript_type "protein_coding"; transcript_id "B.2";
    chr2    test    CDS     75      150     .       +       .       gene_name "B"; transcript_type "protein_coding"; transcript_id "B.2";
    chr2    test    CDS     500     600     .       +       .       gene_name "B"; transcript_type "protein_coding"; transcript_id "B.2";
    chr2    test    CDS     9000    9050    .       +       .       gene_name "B"; transcript_type "protein_coding"; transcript_id "B.2";

    chr3    test    transcript      1       10000   .       +       .       gene_name "C"; transcript_type "protein_coding"; transcript_id "C.1";
    chr3    test    exon    1       150     .       +       .       gene_name "C"; transcript_type "protein_coding"; transcript_id "C.1";
    chr3    test    exon    9000    10000   .       +       .       gene_name "C"; transcript_type "protein_coding"; transcript_id "C.1";
    chr3    test    CDS     75      150     .       +       .       gene_name "C"; transcript_type "protein_coding"; transcript_id "C.2";
    chr3    test    CDS     9000    9050    .       +       .       gene_name "C"; transcript_type "protein_coding"; transcript_id "C.2";
    chr3    test    transcript      1000    2100    .       -       .       gene_name "D"; transcript_type "lincRNA"; transcript_id "D.1";                                                                     
    chr3    test    exon    1000    2100    .       -       .       gene_name "D"; transcript_type "lincRNA"; transcript_id "D.1";
   
    """
    regions = [
        # intergenic
        (('chr1', '+', [(20000, 20010), ]), ("-", "-", "-")),
        (('chrA', '+', [(20000, 20010), ]), ("-", "-", "-")),
        # single gene
        (('chr1', '+', [(0, 74), ]), ("A", "N", "C")),
        (('chr1', '+', [(75, 150), ]), ("A", "C", "C")),
        (('chr1', '+', [(0, 150), ]), ("A", "N,C", "C")),
        (('chr1', '+', [(75, 151), ]), ("A", "C,I", "C")),
        (('chr1', '-', [(0, 150), ]), ("A", "n,c", "C")),
        (('chr1', '+', [(8050, 9040), ]), ("A", "I,C", "C")),
        (('chr1', '+', [(8090, 9051), ]), ("A", "I,C,N", "C")),
        # alternative splicing
        (('chr2', '+', [(450, 550), ]), ("B", "I,C|I", "C")),
        (('chr2', '+', [(140, 150), (500,590)]), ("B", "C|N,C|I", "C")),
        (('chr2', '+', [(9000, 10000)]), ("B", "C|N,N", "C")),
        # multiple genes on opposite strands
        # TODO: we may want to use the information if part of a read aligns 
        # outside of an annotated feature to de-prioritize it
        (('chr3', '+', [(950, 1050), ]), ("C,D", "I,n", "C,I")),
        (('chr3', '+', [(900, 1000), ]), ("C,D", "I,n", "C,I")),
        (('chr3', '-', [(900, 1000), ]), ("C,D", "i,N", "C,I")),
        (('chr3', '*', [(900, 1000), ]), ("C,D", "I,N", "C,I")),

    ]

    ga = GenomeAnnotation.from_compiled_index(test_compile)
    
    fail = []
    for (chrom, strand, blocks), expect in regions:
        res = ga.get_annotation_tags(chrom, strand, blocks)
        gn, gf, gt = res
        print(f'gn "{gn}"; gf "{gf}"; gt "{gt}"; expect={expect}')
        if res != expect:
            fail.append((chrom, strand, blocks, res, expect))
    
    for f in fail:
        print(f"failed test: {f}")

    assert len(fail) == 0





import os


def load(config_file="", load_defaults=True, args={}):
    """
    Tries to load YAML configuration from
        0) builtin default from scbamtools package (if load_defaults=True)
        1) update with yaml loaded from config_file (if provided)
        2) update with kwargs from args (for commandline overrides)
    """
    from yaml import load, Loader
    from argparse import Namespace

    # def recursive_to_namespace(d):
    #     for k, v in d.items():
    #         if type(v) is dict:
    # d[k] = Namespace()
    args_ns = Namespace(**args)

    def recursive_update(dst, src):
        for k, v in src.items():
            if type(v) is dict:
                dst[k] = recursive_update(dst.get(k, {}), v)
            # elif type(v) is list:
            #     dst[k] = dst.get(k, []).extend(v)
            else:
                if type(v) is str and "{" in v:
                    # attempt to expand variables
                    # print(v, args)
                    try:
                        v = v.format(args=args_ns)
                    except KeyError:
                        pass

                dst[k] = v
        return dst

    default_path = os.path.join(os.path.dirname(__file__), "default.yaml")
    default = load(open(default_path), Loader=Loader)
    config = default.copy()

    if config_file:
        try:
            user_config = load(open(config_file), Loader=Loader)
            config = recursive_update(config, user_config)
        except FileNotFoundError:
            pass

    if args:
        config = recursive_update(config, args)

    return Namespace(**config)


def save(config, filename):
    from yaml import dump, Dumper

    open(filename, "wt").write(dump(config))



import numpy as np
import logging
from collections import defaultdict, OrderedDict
from scbamtools.contrib import __version__, __author__, __license__, __email__

# from mrfifo.util import CountingStatistics
import scbamtools.util as util

"""
This module implements the functionality needed to aggregate annotated alignments from a BAM file
into a digital gene expression (DGE) matrix.
The matrix has gene names as columns and cell barcodes as rows. It is stored as a scipy.sparse.csr_matrix
to save RAM. Depending on the kinds of features a read can be annotated with, counting goes through
two stages of disambiguation (see below) and then a determination of "how to count" the read. A read can be 
counted into more than one count-matrix which will be stored as 'layers' in the resulting AnnData object
(see below).

Configuration of the counting rules is done via the config.yaml. All parameters below have a 
corresponding entry there. The section "quant" holds is subdivided by flavors (for example "default") 
and each flavor can set the values of any of the options below. If you choose to customize further,
you can even set a different (sub-)class to be used for counting and potentially define additional paramters
to configure it in the YAML as well.

## BAM ingestion

BAM records are first grouped into 'bundles' of alternative alignments for the same read. Each bundle
consists of a cell barcode, a UMI sequence and a list of one or more possible alignments. For the purpose
of counting gene expression, the different alignments are represented by the values of the gn, gf, and gt 
BAM tags.

## alignment disambiguation

If we have more than one alignment in a bundle, the next step is to try and disambiguate between them.
For this purpose, the alignments are assigned priorites from the 'alignment_priorities' dictionary.
If there is a tie with multiple alignments having the same priority, the disambiguation failed and
the read can not be counted (for example aligning to two paralogous genes with exactly the same alignment
score). However, if there is one alignment with the highest priority, disambiguation has succeeded 
(for example an exonic region of one specific gene being preferred over alternative intergenic 
alignments).

## gene disambiguation

Even if a single alignment can be selected, there can still be conflicting annotations for a read: it 
may align to a region that is annotated for more than one overlapping genes. In this case another round
of disambiguation is attempted, this time using the 'gene_priorities' dictionary.
Here, for example, a coding exon annotation in the sense direction of the read would take precedence
over an intronic annotation to another gene on the opposite strand.

## channel determination

Lastly, if a read has been found to be assignable to one gene, we may still count it into different
channels (note, within each channel, the sum of all counts will never exceed the number of reads
Reads are never counted multiple times into the same channel. That includes the main adata.X channel).

For example, if a read aligns in an intronic region, it can be counted into 'intronic_reads'. If it is
furthermore the first time we observe this combination of cell barcode and UMI, it would additionally be 
counted as 'intronic_counts' ('counts' meaning "unique molecules").
If 'intronic_counts' are included in the "default_X_counts" parameters, it would also count to the main
DGE matrix stored as adata.X .

## A note on parallelization

In order to decide if a read has been seen more than once already (copies of the same BC & UMI sequences)
each DefaultCounter instance maintains a set() of hash values. To ensure correctness but still allow
parallelization, the input process will place bundles into distinct queues based on the cell barcode
sequence.

For example, if we want to run 4 counting-processes in parallel, we would map the first nucleotide to the 
processes such that barcodes starting with 'A'-> process 1, 'C' -> process 2, 'G' -> process 3, 
'T' -> process 3. In this manner the separate sets of hash values are guaranteed to never overlap and the
test for UMI uniqueness is not an issue.
"""

default_alignment_priorities = {
    "C": 100,  # coding exon
    "U": 100,  # UTR exon
    "N": 90,  # exon of non-coding transcript
    "I": 50,  # intronic region
    # lower case == antisense
    "c": 10,
    "u": 10,
    "n": 9,
    "i": 5,
    # intergenic
    "-": 0,
}
default_gene_priorities = {
    "C": 110,  # coding exon
    "U": 100,  # UTR exon
    "N": 90,  # exon of non-coding transcript
    "I": 50,  # intronic region
    "c": 11,
    "u": 10,
    "n": 9,
    "i": 5,
    "-": 0,
}
default_exonic_tags = ["C", "U", "N"]
default_exonic_tags += [t.lower() for t in default_exonic_tags]
default_intronic_tags = ["I", "i"]

default_channels = [
    "counts",
    "reads",
    "exonic_counts",
    "exonic_reads",
    "intronic_counts",
    "intronic_reads",
]
default_X_counts = [
    "exonic_counts",
    "intronic_counts",
]  # what should be counted into adata.X matrix
default_X_reads = [
    "exonic_reads",
    "intronic_reads",
]  # what should be counted into adata.layers["reads"] matrix (correspond with adata.X but for reads not UMIs)


class BaseCounter:

    logger = logging.getLogger("scbamtools.quant.BaseCounter")

    def __init__(
        self,
        channels=default_channels,
        exonic_tags=default_exonic_tags,
        intronic_tags=default_intronic_tags,
        X_counts=default_X_counts,
        X_reads=default_X_reads,
        handle_multimappers=True,
        uniq=set(),
        stats=defaultdict(int),
        **kw,
    ):

        self.kw = kw
        self.filter_sam_records = None  # can be set to a callable which may perform arbitrary filter operations on the SAM columns
        self.channels = channels
        self.exonic_set = set(exonic_tags)
        self.intronic_set = set(intronic_tags)

        # which channels shall contribute to the adata.X (main channel)
        self.count_X_channels = set(X_counts)
        # which channels shall contribute to the adata.layers['reads'] (main channel reads version)
        self.read_X_channels = set(X_reads)
        self.handle_multimappers = handle_multimappers

        self.stats = stats
        self.uniq = uniq

    ## Implement/overload the following functions in a sub-class to implement a desired counting behavior
    def unique_alignment(self, bundle):
        # nothing to do here, we have a unique alignment
        return bundle[0]

    def select_alignment(self, bundle):
        return None

    def select_gene(self, CB, MI, chrom, strand, gn, gf, score):
        if len(gn) == 1:
            return gn[0], gf[0]
        else:
            return "-", "-"

    def exon_intron_disambiguation(self, channels):
        # how to handle reads that align to both intron and exon features
        # default implementation counts everything
        return channels

    ## Count-channel determination
    def determine_channels(self, gf, uniq):
        channels = set()
        exon = False
        intron = False
        for f in gf:
            if f in self.exonic_set:
                channels.add("exonic_reads")
                exon = True
                if uniq:
                    channels.add("exonic_counts")

            elif f in self.intronic_set:
                channels.add("intronic_reads")
                intron = True
                if uniq:
                    channels.add("intronic_counts")

        # post-process: if both exon & intron annotations
        # (but from different isoforms) are present
        # decide how to count
        if exon and intron:
            channels = self.exon_intron_disambiguation(channels)

        if channels & self.count_X_channels:
            channels.add("counts")

        if channels & self.read_X_channels:
            channels.add("reads")

        return channels

    def bundles_from_SAM(self, sam_input):
        import re

        def extract(cols):
            # yield chrom, strand, gn, gf, score
            chrom = cols[2]
            strand = "-" if int(cols[3]) & 16 else "+"
            gn = "-"
            gf = "-"
            AS = -1
            CB = "-"
            MI = "-"

            for c in cols[11:]:
                if c.startswith("gn"):
                    gn = c[5:]
                elif c.startswith("gf"):
                    gf = c[5:]
                elif c.startswith("AS"):
                    AS = int(c[5:])
                elif c.startswith("MI"):
                    MI = c[5:]
                elif c.startswith("CB"):
                    CB = c[5:]

            return CB, MI, chrom, strand, gn, gf, AS

        bundle = []
        last_qname = ""
        for sam in sam_input:
            self.stats["SAM_records_total"] += 1
            cols = sam.rstrip().split("\t")
            if self.filter_sam_records:
                if not self.filter_sam_records(cols):
                    continue

            qname = cols[0]
            if qname != last_qname:
                if bundle:
                    yield bundle

                bundle = []
                last_qname = qname
            bundle.append(extract(cols))

        if bundle:
            yield bundle

    ## main function: alignment bundle -> counting channels
    def process_bundle(self, bundle):
        self.stats["SAM_bundles_total"] += 1
        gene = None
        selected = None
        channels = set()
        # self.set_reference(reference_name)
        # print(f"bundle={bundle}")
        if len(bundle) == 1:
            self.stats["N_aln_unique"] += 1
            selected = self.unique_alignment(bundle)
        elif self.handle_multimappers:
            self.stats["N_aln_multi"] += 1
            selected = self.select_alignment(bundle)
            if selected:
                self.stats["N_aln_selected"] += 1
            else:
                self.stats["N_aln_selection_failed"] += 1
        # print(f"selected -> {selected}")
        if selected:
            self.stats["N_aln_countable"] += 1
            cell, umi, chrom, strand, gn, gf, score = selected
            # self.stats[f'N_aln_{chrom}'] += 1
            self.stats[f"N_aln_{strand}"] += 1

            gene, gf = self.select_gene(*selected)
            # print(f"gene, gf -> {gene}, {gf}")
            if len(gn) == 1:
                gene = gn[0]
                if gene is None or gene == "-":
                    self.stats["N_gene_none"] += 1
                else:
                    self.stats["N_gene_unique"] += 1
                # gf = gf
                # print(f"uniq gene: {gene} {gf}")
            else:
                self.stats["N_gene_multi"] += 1
                if gene:
                    self.stats["N_gene_selected"] += 1
                else:
                    self.stats["N_gene_selection_failed"] += 1

            # count the alignment every way that the counter prescribes
            if gene != "-":
                # print(f"gene={gene} gf={gf}")
                self.stats["N_aln_counted"] += 1
                if gf[0].islower():
                    self.stats["N_aln_antisense"] += 1
                else:
                    self.stats["N_aln_sense"] += 1

                # handle the whole uniqueness in a way that can parallelize
                # maybe split by UMI[:2]? This way, distributed uniq() sets would
                # never overlap
                key = hash((cell, gene, umi))
                uniq = not (key in self.uniq)
                if uniq:
                    self.uniq.add(key)

                channels = self.determine_channels(gf, uniq)
                for c in channels:
                    self.stats[f"N_channel_{c}"] += 1

                if not channels:
                    self.stats[f"N_channel_NONE"] += 1
        else:
            cell = "-"
            gene = "-"
            channels = set()

        return cell, gene, channels


class CustomIndexCounter(BaseCounter):

    logger = logging.getLogger("scbamtools.quant.CustomIndexCounter")

    def unique_alignment(self, bundle):
        CB, MI, chrom, strand, _, _, score = bundle[0]
        return (CB, MI, chrom, strand, [chrom], ["N"], score)

    def select_alignment(self, bundle):
        # only consider alignments on the + strand (for custom indices)
        plus = [b for b in bundle if b[1] == "+"]
        if plus:
            return self.unique_alignment(plus)

    # def determine_channels(self, gf, uniq):
    #     if uniq:
    #         channels = {'reads', 'counts'}
    #     else:
    #         channels = {'reads'}

    #     return channels

    # select_gene and exon_intron_disambiguation never get called
    # and thus do not need to be implemented


class mRNACounter(BaseCounter):

    logger = logging.getLogger("scbamtools.quant.mRNACounter")

    def __init__(
        self,
        alignment_priorities=default_alignment_priorities,
        gene_priorities=default_gene_priorities,
        **kw,
    ):

        BaseCounter.__init__(self, **kw)

        self.alignment_priorities = alignment_priorities
        self.gene_priorities = gene_priorities

    def unique_alignment(self, bundle):
        # no processing needed here. We have a unique alignment
        return bundle[0]

    ## Try to select the most meaningful (cDNA-derived) from multiple
    ## reported alignments for a read
    def select_alignment(self, bundle):
        """
        If multiple alignments are reported for one fragment/read try to make a reasonable choice:
        If one of the alignments is to a coding exon, but the others are not, choose the exonic one.
        If multiple alternative alignments point to exons of different coding genes, do not count anything
        because we can not disambiguate.
        """
        from collections import defaultdict

        def get_prio(gf):
            prios = []
            for f in gf:
                f_prios = [self.gene_priorities.get(x, 0) for x in f.split("|")]
                prios.append(max(f_prios))

            # prios = [self.alignment_priorities.get(f, 0) for f in gf]
            return max(prios)

        top_prio = 0
        n_top = 0
        kept = None

        for CB, MI, chrom, strand, gn, gf, score in bundle:
            p = get_prio(gf)
            if p > top_prio:
                top_prio = p
                kept = (CB, MI, chrom, strand, gn, gf, score)
                n_top = 1
            elif p == top_prio:
                n_top += 1
                kept = (CB, MI, chrom, strand, gn, gf, score)

        if n_top == 1:
            # print(kept)
            return kept

    ## Gene selection strategy, similar to alignment selection
    def select_gene(self, CB, MI, chrom, strand, gn, gf, score):
        # let's see if we can prioritize which gene we are interested in
        gene_prio = defaultdict(int)
        gene_gf = defaultdict(set)
        max_prio = 0
        max_code = "-"
        import itertools

        gn = gn.split(",")
        gf = gf.split(",")
        for n, f in itertools.zip_longest(gn, gf, fillvalue=gn[0]):
            codes = f.split("|")
            f_prios = np.array([self.gene_priorities.get(x, 0) for x in codes])
            i = f_prios.argmax()
            p = f_prios[i]
            c = codes[i]
            gene_gf[n].add(c)

            if p < 0:
                gene_prio[n] -= p
                # we get a penalty. Useful if the overlap extends
                # beyond the boundaries of a feature
            elif gene_prio[n] <= p:
                # we have found are higher priority annotation
                gene_prio[n] = p
                # keep only the highest priority codes per compound code,
                # examples:
                # C|I -> C
                # N|U|I -> U
                max_prio = max([max_prio, p])

        # restrict to genes tied for highest priority hits
        gn_new = [n for n in set(gn) if gene_prio[n] == max_prio]
        # print(len(gn_new))
        if len(gn_new) == 1:
            # YES we can salvage this read
            gene = gn_new[0]
            res = gene, sorted(gene_gf[gene])
        else:
            # NO still ambiguous
            res = "-", "-"

        # print(f"gn={gn} gf={gf} -> res={res}")
        return res

    def exon_intron_disambiguation(self, channels):
        return channels - set(["exonic_reads", "exonic_counts"])


class SLAC_miRNACounter(BaseCounter):

    logger = logging.getLogger("scbamtools.quant.SLAC_miRNACounter")

    def __init__(
        self,
        *argc,
        max_5p_soft_clip=5,
        max_3p_soft_clip=None,
        min_matches=15,
        max_pos=3,
        **kwargs,
    ):
        super().__init__(self, *argc, **kwargs)
        self.max_5p_soft_clip = max_5p_soft_clip
        self.max_3p_soft_clip = max_3p_soft_clip
        self.min_matches = min_matches
        self.max_pos = max_pos
        self.filter_sam_records = self.cigar_filter

    def cigar_filter(self, cols):
        import re

        if self.max_pos is not None:
            pos = int(cols[3])
            if pos > self.max_pos:
                return False

        matches = 0
        cigar = cols[5]
        cigar_ops = re.findall(r"(\d+)(\w)", cigar)

        # check 5' most operation
        n, o = cigar_ops.pop(0)
        n = int(n)
        if o == "S":
            if n > self.max_5p_soft_clip:
                self.stats["CIGAR_excessive_5p_clipping"] += 1
                return False
        elif o == "M":
            matches += n

        if cigar_ops and (self.max_3p_soft_clip is not None):
            # check 3' most operation
            n, o = cigar_ops.pop(-1)
            n = int(n)
            if o == "S":
                if n > self.max_3p_soft_clip:
                    self.stats["CIGAR_excessive_3p_clipping"] += 1
                    return False
            elif o == "M":
                # assume it is 'M'
                matches += n

        if self.min_matches is not None:
            if cigar_ops:
                for n, o in cigar_ops:
                    if o == "M":
                        matches += int(n)

            if matches <= self.min_matches:
                self.stats["CIGAR_insufficient_match"] += 1
                return False

        self.stats["CIGAR_pass"] += 1
        return True

    def unique_alignment(self, bundle):
        CB, MI, chrom, strand, _, _, score = bundle[0]
        return (CB, MI, chrom, strand, [chrom], ["N"], score)

    def select_alignment(self, bundle):
        # only consider alignments on the + strand (for custom indices)
        best_score = 0
        best_score_aln = None
        best_score_tie = 0

        for aln in bundle:
            CB, MI, chrom, strand, gn, gf, score = aln
            if strand == "+":
                if score > best_score:
                    best_score = score
                    best_score_aln = CB, MI, chrom, strand, [chrom], ["N"], score
                    best_score_tie = 1
                elif score == best_score:
                    best_score_tie += 1

        if best_score_tie == 1:
            return best_score_aln

    def select_gene(self, CB, MI, chrom, strand, gn, gf, score):
        # print(CB, MI, chrom, gn, gf)
        return [chrom], ["N"]


DefaultCounter = mRNACounter


def sparse_summation(X, axis=0):
    # for csr_array this would be fine
    # return np.array(X.sum(axis=axis))
    # unfortunately, for the time being scanpy needs csr_matrix
    # due to h5py lagging somewhat
    if axis == 0:
        return np.array(X.sum(axis=0))[0]
    elif axis == 1:
        return np.array(X.sum(axis=1))[:, 0]


class DGE:
    logger = logging.getLogger("scbamtools.quant.DGE")

    def __init__(self, channels=["count"], cell_bc_allow_list=""):
        self.channels = channels
        self.counts = defaultdict(lambda: defaultdict(int))
        self.DGE_cells = set()
        self.DGE_genes = set()
        self.allowed_bcs = self.load_list(cell_bc_allow_list)

    @staticmethod
    def load_list(fname):
        if fname:
            allowed_bcs = set([line.rstrip() for line in open(fname)])
            DGE.logger.debug(f"restricting to {len(allowed_bcs)} allowed cell barcodes")
            return allowed_bcs
        else:
            DGE.logger.debug("all barcodes allowed")

    def add_read(self, cell, gene, channels=["count"]):
        if channels:
            if self.allowed_bcs and not (cell in self.allowed_bcs):
                # print(f"discarding '{cell}' as NA", cell in self.allowed_bcs)
                cell = "NA"

            self.DGE_cells.add(cell)
            self.DGE_genes.add(gene)

            d = self.counts[(gene, cell)]
            for channel in channels:
                d[channel] += 1

    def make_sparse_arrays(self):
        """_summary_
        Convert counts across all channels into a dictionary of sparse arrays

        Returns:
            dict: channel is key, value is scipt.sparse.csr_array
        """
        import scipy.sparse
        import anndata

        ind_of_cell = {}
        obs = []
        for cell in sorted(self.DGE_cells):
            ind_of_cell[cell] = len(obs)
            obs.append(cell)

        ind_of_gene = {}
        var = []
        for gene in sorted(self.DGE_genes):
            ind_of_gene[gene] = len(var)
            var.append(gene)

        counts_by_channel = defaultdict(list)
        row_ind = []
        col_ind = []
        for (gene, cell), channel_dict in self.counts.items():
            for channel in self.channels:
                counts_by_channel[channel].append(channel_dict[channel])

            row_ind.append(ind_of_cell[cell])
            col_ind.append(ind_of_gene[gene])

        if (not len(row_ind)) or (not len(col_ind)):
            self.logger.warning(
                f"empty indices len(row_ind)={len(row_ind)} len(col_ind)={len(col_ind)}"
            )
            return {}, [], []

        sparse_channels = OrderedDict()
        for channel in self.channels:
            counts = counts_by_channel[channel]
            self.logger.debug(
                f"constructing sparse-matrix for channel '{channel}' from n={len(counts)} non-zero entries (sum={np.array(counts).sum()})"
            )
            m = scipy.sparse.csr_matrix((counts, (row_ind, col_ind)), dtype=np.float32)
            self.logger.debug(
                f"resulting sparse matrix shape={m.shape} len(obs)={len(obs)} len(var)={len(var)}"
            )
            sparse_channels[channel] = m

        return sparse_channels, obs, var

    @staticmethod
    def sparse_arrays_to_adata(channel_d, obs, var, main_channel="counts"):
        """_summary_
        Creates an AnnData object from sparse matrices in the channel_d dictionary. The main_channel
        will be assigned to adata.X, the others to adata.layers[channel].

            obs (cell barcodes) -> rows
            var (gene names) -> columns

        Args:
            channel_d (_type_): _description_
            obs (_type_): _description_
            var (_type_): _description_
            main_channel (_type_): _description_

        Returns:
            AnnData object: a proper, scanpy-compatible AnnData object with (optional) layers.
        """
        import anndata

        X = channel_d[main_channel]
        adata = anndata.AnnData(X)
        adata.obs_names = obs
        adata.var_names = var

        for channel, sparse in channel_d.items():
            if channel != main_channel:
                adata.layers[channel] = sparse
                # add marginal counts
                adata.obs[f"n_{channel}"] = sparse_summation(
                    adata.layers[channel], axis=1
                )

        # number of molecules counted for this cell
        adata.obs[f"n_counts"] = sparse_summation(adata.X, axis=1)
        # number of genes detected in this cell
        adata.obs["n_genes"] = sparse_summation(adata.X > 0, axis=1)
        return adata



import pytest
import scbamtools.quant as quant
import os

test_bundles = [
    ("empty", [], ("-", "-", set())),
    (
        "uniq",
        [("ACGTACGT", "AAAAC", "chr1", "+", "A", "N", 0)],
        ("A", {"counts", "reads", "exonic_counts", "exonic_reads"}),
    ),
    (
        "mm_gene_intergenic",
        [
            ("ACGTACGT", "AAAAG", "chr1", "+", "A", "C", 0),
            ("ACGTACGT", "AAAAG", "chr2", "-", "-", "-", 0),
        ],
        ("ACGTACGT", "A", {"counts", "reads", "exonic_counts", "exonic_reads"}),
    ),
    (
        "mm_geneC_geneI",
        [
            ("ACGTACGT", "AAACA", "chr1", "+", "A", "C", 0),
            ("ACGTACGT", "AAACA", "chr2", "-", "B", "I", 0),
        ],
        ("ACGTACGT", "A", {"counts", "reads", "exonic_counts", "exonic_reads"}),
    ),
    (
        "mm_complex",
        [
            ("ACGTACGT", "ACACA", "chr1", "+", "A", "C|N|I,I", 0),
            ("ACGTACGT", "ACACA", "chr2", "-", "B", "n", 0),
        ],
        ("ACGTACGT", "A", {"counts", "reads", "intronic_counts", "intronic_reads"}),
    ),
    (
        "mm_complex2",
        [
            ("ACGTACGT", "ATACA", "chr1", "+", "A", "N,C|N|I,U", 0),
            ("ACGTACGT", "ATACA", "chr2", "-", "B", "n", 0),
        ],
        ("ACGTACGT", "A", {"counts", "reads", "exonic_counts", "exonic_reads"}),
    ),
    (
        "mm_ambig",
        [
            ("ACGTACGT", "AAACG", "chr1", "+", "A", "C", 0),
            ("ACGTACGT", "AAACG", "chr2", "-", "B", "C", 0),
        ],
        ("ACGTACGT", None, set()),
    ),
    (
        "mm_geneI_geneN",
        [
            ("ACGTACGT", "AAACT", "chr1", "+", "A,B", "I,n", 0),
        ],
        ("ACGTACGT", "A", {"counts", "reads", "intronic_counts", "intronic_reads"}),
    ),
    (
        "uniq_dup",
        [("ACGTACGT", "AAAAC", "chr1", "+", "A", "N", 0)],
        ("ACGTACGT", "A", {"reads", "exonic_reads"}),
    ),
    (
        "uniq",
        [("NNNNNNNN", "AAAAC", "chr1", "+", "A", "N", 0)],
        ("NNNNNNNN", "A", {"counts", "reads", "exonic_counts", "exonic_reads"}),
    ),
]


def test_gene_selection():
    counter = quant.DefaultCounter()
    assert counter.select_gene("NA", "NA", "chr1", "+", "A,B", "C|U|I,i|n", 90) == (
        "A",
        ["C"],
    )
    assert counter.select_gene("NA", "NA", "chr1", "+", "A", "C|I,I,U", 90) == (
        "A",
        ["C", "I", "U"],
    )


def test_default_counter():
    counter = quant.DefaultCounter()
    sm_dir = os.path.dirname(__file__)
    dge = quant.DGE(
        channels=counter.channels,
        cell_bc_allow_list=sm_dir + "/../test_data/simple_allowlist.txt",
    )

    fail = []
    for bla in test_bundles:
        assert len(bla) == 3
        name, bundle, expect = bla
        res = counter.process_bundle(bundle=bundle)
        cb, gene, channels = res
        print(res == expect, res, expect)
        if res != expect:
            fail.append((name, bundle, res, expect))

        dge.add_read(gene, cb, channels)

    channel_d, obs, var = dge.make_sparse_arrays()

    adata = quant.DGE.sparse_arrays_to_adata(channel_d, obs, var)
    print(adata)



from setuptools import setup, Extension
from Cython.Build import cythonize
from Cython.Compiler import Options
import numpy

# These are optional
Options.docstrings = True
Options.annotate = False

# Modules to be compiled and include_dirs when necessary
extensions = [
    Extension(
        "scbamtools.cython.kmers",
        ["scbamtools/cython/kmers.pyx"],
        include_dirs=[numpy.get_include()],
    ),
    Extension(
        "scbamtools.cython.bctree",
        ["scbamtools/cython/bctree.pyx"],
        include_dirs=[numpy.get_include()],
    ),
]

if __name__ == "__main__":
    from setuptools import setup

    setup(
        name="scbamtools",  # Required
        packages=[
            "scbamtools",
            "scbamtools.cython",
            "scbamtools.bin",
            "scbamtools.config",
        ],
        package_data={"scbamtools/cython": ["kmers.pyx", "bctree.pyx", "types.pxd"]},
        # A list of compiler Directives is available at
        # https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#compiler-directives
        # external to be compiled
        ext_modules=cythonize(
            extensions, compiler_directives={"language_level": 3, "profile": False}
        ),
    )
